"""
get_cfs.py
----------

Script which parses costfunctions.py to see what cost functions are available.
Used by the frontend.

SPDX-License-Identifier: GPL-3.0-or-later
"""

import sys
import ast
import json
from pathlib import Path


def main():
    """
    Parses costfunctions.py to get the top-level functions. Prints them to
    stdout with a space as the delimiter. This allows the frontend to read them
    in using Popen.communicate().
    """
    def get_cfs(file):
        # Gets cost functions and docstrings from a given file.
        # Returns dictionary in the format {'function_name' : 'docstring'}.
        with open(file, 'r') as fp:
            tree = ast.parse(fp.read())
        # Find out which nodes are actually functions.
        functions = [func for func in tree.body
                     if isinstance(func, ast.FunctionDef)]
        # Get their names as well as their docstrings.
        function_names = [func.name for func in functions]
        docstrings = [ast.get_docstring(func) for func in functions]
        return dict(zip(function_names, docstrings))

    cf_system_file = Path(__file__).resolve().parent / "costfunctions.py"
    cf_user_file = Path(__file__).resolve().parent / "costfunctions_user.py"

    # System and user cost functions.
    system_cfs = get_cfs(cf_system_file)
    user_cfs = get_cfs(cf_user_file)
    # {**x, **y} causes the entries from y to overwrite entries from x with the
    # same name, which means that user cost functions shadow system cost
    # functions. In 3.9+ we can do this with all_cfs = system_cfs | user_cfs :)
    all_cfs = {**system_cfs, **user_cfs}
    print(json.dumps(all_cfs))


if __name__ == "__main__":
    main()



"""
costfunctions.py
----------------

Contains default cost functions. Custom cost functions should be added in
costfunctions_user.py (using the same format).

You can add custom cost functions in here (they will work), but they will be
shadowed by any similarly named cost functions in costfunctions_user.py. Also,
this file will be overwritten if POISE is reinstalled.

SPDX-License-Identifier: GPL-3.0-or-later
"""

import numpy as np

from .cfhelpers import *
from .shared import _g


def noe_1d():
    """
    Measures the intensity of peaks in the spectrum, *except* for anything
    within 25 Hz of the selectively excited peak.
    """
    # SPOFFS2 plus O1 is the frequency of the selective pulse, in Hz.
    # We can get the values of both using getpar(), which automatically
    # converts numeric parameters to floats.
    f = getpar("SPOFFS2") + getpar("O1")
    # Define the (rough) bandwidth of the selective pulse. We will ignore the
    # region of the spectrum that is bw Hz wide and centred on the selective
    # pulse.
    bw = 50.0
    # To get the spectrum *without* the ignored region, we split up the
    # spectrum into two "portions". We need to construct the 'bounds' parameter
    # for both portions.
    sfo1 = getpar("SFO1")
    lowerhalf = f"{(f + bw/2)/sfo1:.3f}.."   # frequencies (f + bw)/2 and above
    upperhalf = f"..{(f - bw/2)/sfo1:.3f}"   # frequencies up to (f - bw)/2
    # We use get1d_real() to get both portions of the spectrum, passing the
    # bounds as appropriate. The intensities of the spectrum are obtained by
    # summing all points (np.sum).
    # Note that manually passing the bounds parameter will override the F1P/F2P
    # parameters if set by the user (this is probably a good thing).
    upper_integral = np.sum(get1d_real(bounds=upperhalf))
    lower_integral = np.sum(get1d_real(bounds=lowerhalf))
    # Lastly, we take the negative absolute value of the resulting sum, in
    # order to make sure that the resulting cost function is negative
    # regardless of how the NOE peaks are phased (positive or negative).
    return -abs(upper_integral + lower_integral)


def minabsint():
    """
    Cost function which minimises the absolute (magnitude-mode) intensity of
    the spectrum. This is probably the easiest cost function. :-)
    """
    # Get the real part of the spectrum as a numpy ndarray. Because no bounds
    # are explicitly passed, this defaults to the whole spectrum if F1P/F2P are
    # not specified. If F1P/F2P *are* specified (either via the `dpl` TopSpin
    # command, or manually), this will only return the part of the spectrum
    # between those two values.
    r = get1d_real()
    # ...and the imaginary part
    i = get1d_imag()
    # The magnitude-mode spectrum is obtained by combining these.
    mag = np.abs(r + 1j * i)
    # The intensity of the magnitude-mode spectrum is just the sum of all
    # points.
    return np.sum(mag)


def maxabsint():
    """
    Cost function which maximises the absolute (magnitude-mode) intensity of
    the spectrum.
    """
    r = get1d_real()
    i = get1d_imag()
    # This is the same as minabsint except that we have a negative sign.
    # Because the optimisation always seeks to *minimise* the cost function,
    # this essentially tries to *maximise* np.sum(...), i.e. maximise the
    # spectral intensity.
    return -np.sum(np.abs(r + 1j*i))


def minrealint():
    """
    Minimises the intensity of the real part of the spectrum. If the spectrum
    has negative peaks this cost function will try to maximise those.
    """
    return np.sum(get1d_real())


def maxrealint():
    """
    Maximises the intensity of the real part of the spectrum.
    """
    return -np.sum(get1d_real())


def zerorealint():
    """
    Tries to get the intensity of the real spectrum to be as close to zero as
    possible. This works by summation, so dispersion-mode peaks will not
    contribute to this cost function (as they add to zero).
    """
    return np.abs(np.sum(get1d_real()))


def dosy():
    """
    Tries to make a spectrum (1D diffusion experiment with variable gradient
    strength) have 25% intensity of a target spectrum (1D diffusion experiment
    with minimum gradient strength). This should be used in conjunction with
    the dosy_opt Python programme (after POISE is installed via `pip`, this can
    be installed from the TopSpin command-line via `poise --install dosy`.)
    """
    # The target spectrum is hardcoded to have EXPNO 99998. See psyche() for
    # more discussion about this.
    # As in psyche(), the F1P/F2P parameters are not taken from the reference
    # spectrum but from the spectrum being optimised.
    reference_path = make_p_spec(expno=99998, procno=1)
    target = get1d_real(p_spec=reference_path)
    # This is the spectrum being optimised.
    spec = get1d_real()
    # The intensity of the optimised spectrum is np.sum(spec), and likewise for
    # the target spectrum. We take the ratio of the two and measure how "far
    # away" this is from the ideal value of 0.25 (which corresponds to 75%
    # attenuation).
    return np.abs(np.sum(spec)/np.sum(target) - 0.25)


def dosy_aux():
    """
    Non-absolute value of dosy(). To be used in the first stage of dosy_opt.
    """
    reference_path = make_p_spec(expno=99998, procno=1)
    target = get1d_real(p_spec=reference_path)
    spec = get1d_real()
    return np.sum(spec)/np.sum(target) - 0.25


# The cost functions below were used in the POISE paper, but are less likely to
# be generally useful. If you want to use them, simply uncomment them (remove
# the three single quotes around them).

'''
def asaphsqc():
    """
    Maximises the intensity of the projection of a 2D spectrum onto the f2
    axis.
    """
    # First we get the 'rr' part of the spectrum (the displayed part). This
    # assumes that it has been phased already (by the AU program).
    # This is a 2D numpy array:
    # [ [1  , 2, 3, ..., si2],
    #   [.  , ., ., ..., ...],
    #   [.  , ., ., ..., ...],
    #   [si1, ., ., ..., ...]]
    # where si1 and si2 are the values of SI in the indirect and direct
    # dimension respectively.
    spec = get2d_rr()
    # We take the 'skyline' projection, i.e. the highest point along each
    # column. This is basically equivalent to TopSpin's `f2projp` command.
    # If you want a projection onto f1 instead of f2, you can use axis=1.
    proj = np.amax(spec, axis=0)
    return -np.sum(proj)
'''

'''
def psyche():
    """
    Returns the "spectral difference" between the current spectrum and a
    reference spectrum with EXPNO 1 and PROCNO 1.
    """
    # We can construct the path to the reference spectrum using make_p_spec().
    reference_path = make_p_spec(expno=1, procno=1)
    # The way this is currently coded, the reference spectrum has to be placed
    # in EXPNO 1 and PROCNO 1. In theory, this can be specified using a TopSpin
    # parameter in the optimisation dataset, for example "cnst30" (or any other
    # random cnst). In order to use this, just add:
    #     expno = # int(getpar("cnst30"))
    # and then after that you can use
    #     make_p_spec(expno=expno, procno=1)
    # Then we can get the reference, or 'target', spectrum as a numpy array.
    # Note that this will respect the F1P/F2P parameters, *not* of the
    # reference spectrum, but of the spectrum being optimised. This is because
    # get1d_real() does not actually read F1P and F2P from the spectrum. It
    # just uses the global values stored in _g.spec_f1p and _g.spec_f2p, which
    # refer to the F1P/F2P values of the spectrum being optimised.
    target = get1d_real(p_spec=reference_path)
    # ... and the current spectrum that is being optimised.
    spec = get1d_real()
    # Finally, return the 2-norm of the difference between the two normalised
    # spectra.
    return np.linalg.norm(target/np.linalg.norm(target) -
                          spec/np.linalg.norm(spec))
'''


'''
def dosy_2p():
    """
    Two-parameter DOSY optimisation.
    """
    reference_path = make_p_spec(expno=99998, procno=1)
    target = get1d_real(p_spec=reference_path)
    spec = get1d_real()
    return np.abs(np.sum(spec)/np.sum(target) - 0.25) + getpar("D20")
'''



"""
optpoise.py
-----------

Module containing optimisation functions.

SPDX-License-Identifier: GPL-3.0-or-later
"""

from .cfhelpers import CostFunctionError

from functools import wraps

import numpy as np
import pybobyqa as pb

# Magic constant which tells us how much to scale each tolerance to when
# scaling by tols. It is basically arbitrarily chosen, but should not have an
# effect on the actual optimisation.
MAGIC_TOL = 0.03

# Constant strings denoting "standard" optimisation outcomes.
MESSAGE_OPT_SUCCESS = "Optimisation terminated successfully due to convergence."
MESSAGE_OPT_PREMATURE_TERMINATION = "Optimisation terminated before convergence."
MESSAGE_OPT_MAXFEV_REACHED = "Maximum function evaluations reached."
MESSAGE_OPT_MAXITER_REACHED = "Maximum iterations reached."


def scale(val, lb, ub, tol, scaleby="bounds"):
    """
    Scales a set of values so that the optimisation behaves better.

    For scaleby="bounds", scales a set of values such that the lower and upper
    bounds for all variables are 0 and 1 respectively). We used to use this.
    As of the current version it's no longer being used.

    For scaleby="tols", scales a set of values such that the tolerances for all
    variables are MAGIC_TOL.

    Parameters
    ----------
    val : ndarray or list
        The unscaled values.
    lb : ndarray or list
        The unscaled lower bounds.
    ub : ndarray or list
        The unscaled upper bounds.
    tol : ndarray or list
        The unscaled tolerances.
    scaleby : str from {"bounds", "tols"}
        Method to scale by.

    Returns
    -------
    scaled_val : ndarray
        The scaled values.
    scaled_lb : ndarray
        The scaled lower bounds.
    scaled_ub : ndarray
        The scaled upper bounds.
    scaled_tol : ndarray
        The scaled tolerances.
    """
    if scaleby not in ["bounds", "tols"]:
        raise ValueError(f"Invalid argument scaleby={scaleby} given.")
    # Convert to ndarray
    val, lb, ub, tol = (np.array(i) for i in (val, lb, ub, tol))
    # Check if any are outside bounds
    if np.any(val < lb) or np.any(val > ub):
        return None
    # Scale them
    if scaleby == "bounds":
        scaled_val = (val - lb)/(ub - lb)
        scaled_lb = (lb - lb)/(ub - lb)  # all 0's
        scaled_ub = (ub - lb)/(ub - lb)  # all 1's
        scaled_tol = tol/(ub - lb)
    elif scaleby == "tols":
        scaled_val = (val - lb) * MAGIC_TOL / tol
        scaled_lb = (lb - lb) * MAGIC_TOL / tol   # all 0's
        scaled_ub = (ub - lb) * MAGIC_TOL / tol
        scaled_tol = tol * MAGIC_TOL / tol        # all MAGIC_TOL's
    return scaled_val, scaled_lb, scaled_ub, scaled_tol


def unscale(scaled_val, orig_lb, orig_ub, orig_tol, scaleby="bounds"):
    """
    Unscales a set of scaled values to their original values.

    Parameters
    ----------
    scaled_val : ndarray or list
        The scaled values.
    orig_lb : ndarray or list
        The *unscaled* lower bounds.
    orig_ub : ndarray or list
        The *unscaled* upper bounds.
    orig_tol : ndarray or list
        The *unscaled* tolerances.
    scaleby : str from {"bounds", "tols"}
        Method to scale by.

    Returns
    -------
    ndarray
        The unscaled values.
    """
    if scaleby not in ["bounds", "tols"]:
        raise ValueError(f"Invalid argument scaleby={scaleby} given.")
    scaled_val, orig_lb, orig_ub, orig_tol = (np.array(i) for i in (scaled_val,
                                                                    orig_lb,
                                                                    orig_ub,
                                                                    orig_tol))
    if scaleby == "bounds":
        return orig_lb + (scaled_val * (orig_ub - orig_lb))
    elif scaleby == "tols":
        return (scaled_val * orig_tol / MAGIC_TOL) + orig_lb


class Simplex():
    """
    Simplex class.
    """
    def __init__(self, x0, method="spendley", length=MAGIC_TOL * 10):
        """
        Initialises a Simplex object.

        Parameters
        ----------
        x0: ndarray
            Initial guess.
        method : str, optional
            Method for construction of initial simplex. Options are:
                - "spendley" : Regular simplex of Spendley et al. (1962) (DOI
                               10.1080/00401706.1962.10490033). This is the
                               default, because it works well.
                - "axis"     : Simplex extended along each axis by a fixed
                               length from x0.
                - "random"   : Every point is randomly generated in [0,1].
                               Don't recommend using this.
        length : float, optional
            A measure of the size of the initial simplex. Defaults to
            MAGIC_TOL * 10, because that's 10 times the tolerances after the
            problem is scaled by tols.
        """
        self.x0 = np.ravel(np.asfarray(x0))
        self.N = np.size(self.x0)

        # Generate simplex
        self.x = np.zeros((self.N + 1, self.N))
        self.f = np.full(self.N + 1, fill_value=np.inf)

        if method == "spendley":
            # Default method.
            # Spendley (1962). DOI 10.1080/00401706.1962.10490033
            # Rosenbrock with x0 = [1.3, 0.7, 0.8, 1.9, 1.2]: 472 nfev, 284 nit
            p = (1/(self.N * np.sqrt(2))) * (self.N - 1 + np.sqrt(self.N + 1))
            q = (1/(self.N * np.sqrt(2))) * (np.sqrt(self.N + 1) - 1)
            self.x[0] = self.x0
            for i in range(1, self.N + 1):
                for j in range(self.N):
                    self.x[i, j] = self.x0[j] + length*p \
                        if j == i - 1 else self.x0[j] + length*q
        elif method == "axis":
            # Axis-by-axis simplex. Each point is just x0 extended along
            # a different axis.
            # Rosenbrock with x0 = [1.3, 0.7, 0.8, 1.9, 1.2]: 566 nfev, 342 nit
            self.x[0] = self.x0
            for i in range(1, self.N + 1):
                for j in range(self.N):
                    self.x[i, j] = self.x0[j] + length \
                        if j == i - 1 else self.x0[j]
        elif method == "random":
            # Every point except x0 is random.
            # Rosenbrock with x0 = [1.3, 0.7, 0.8, 1.9, 1.2]: 705 nfev, 431 nit
            # (average over 1000 iterations)
            self.x[0] = self.x0
            rng = np.random.default_rng()
            for i in range(1, self.N + 1):
                self.x[i] = rng.uniform(size=self.N)
        else:
            raise ValueError(f"invalid simplex generation method "
                             "'{method}' specified")

    def sort(self):
        """
        Sorts the simplex and associated function values in ascending order of
        the cost function, i.e. sim.x[0] contains the current best point,
        sim.x[N] contains the current worst point.
        """
        indices = np.argsort(self.f)
        self.x = np.take(self.x, indices, 0)
        self.f = np.take(self.f, indices, 0)

    def xbar(self):
        """
        Calculates the centroid. Assumes the function values are already
        sorted.
        """
        return np.average(self.x[0:self.N], axis=0)

    def xworst(self):
        """
        Worst point. Assumes the simplex is already sorted.
        """
        return self.x[self.N]

    def replace_worst(self, xnew, fnew):
        """
        Replaces the worst point with the new point xnew, and the corresponding
        function value fnew.
        """
        self.sort()
        self.x[self.N], self.f[self.N] = xnew, fnew

    def shrink(self):
        """
        Performs shrink step (Step 3(f) in Algorithm 8.1.1, Kelley). Doesn't
        evaluate cost functions, only replaces the points!
        """
        for i in range(1, self.N + 1):
            self.x[i] = self.x[0] - (self.x[i] - self.x[0])/2


# Custom exceptions.
class MaxFevalsReached(Exception):
    pass


class MaxItersReached(Exception):
    pass


class EarlyTerminationError(Exception):
    pass


def deco_count(fn):
    """
    Decorator which counts the number of times a function has been called, as
    long as the function does not return np.inf. This makes sure that
    acquire_nmr.calls is not incremented when an out-of-bounds value is
    "sampled".
    """
    @wraps(fn)
    def counter(*args, **kwargs):
        result = fn(*args, **kwargs)
        if result != np.inf:
            counter.calls += 1
        return result
    counter.calls = 0
    return counter


def deco_cf(maxfev):
    """
    Decorator factory which returns a decorator for cost functions. The
    decorator in turn causes the cost function to raise MaxFevalsReached if its
    number of calls is greater than ``maxfev``.

    Usage
    =====
    The decorator factory itself must take a parameter ``maxfev`` so that the
    maximum number of function evaluations can be dynamically chosen. Thus,

    >>> deco_cf(maxfev=maxfev)

    returns a decorator, which can then be used to decorate a cost function:

    >>> cf = deco_cf(maxfev=maxfev)(cf)
    """
    def decorator(fn):
        @wraps(fn)
        def decorated_cf(*args, **kwargs):
            # Check if maximum function evaluations have been reached.
            if fn.calls >= maxfev:
                raise MaxFevalsReached
            # If not, then we can try to run the original function. The
            # parameters returned are the cost function value, a flag
            # indicating whether to break, and a message.
            else:
                return fn(*args, **kwargs)
        return decorated_cf
    return decorator


class OptResult:
    """
    A *very* generic class that exists solely to store the result of an
    optimisation. There's no functionality needed because this is just used
    as a way of passing a message back to the calling script.
    """
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __repr__(self):
        return str(self.__dict__)


def nelder_mead(cf, x0, xtol, scaled_lb, scaled_ub,
                args=(), maxfev=0, simplex_method="spendley"):
    """
    Nelder-Mead optimiser, as described in Section 8.1 of Kelley, "Iterative
    Methods for Optimization".

    Parameters
    ----------
    cf : function
        The cost function. For POISE, this means acquire_nmr(), not the
        user-defined cost function. However in general, this can be any cost
        function. The cost function *must* be decorated with deco_count() (for
        POISE, this is already done).
    x0 : ndarray or list
        Initial point for optimisation. This should already be scaled.
    xtol : ndarray or list
        Tolerances for each optimisation dimension. This should already be
        scaled.
    scaled_lb : ndarray
        Scaled lower bounds for the optimisation.
    scaled_ub : ndarray
        Scaled upper bounds for the optimisation. This is used to place an
        upper bound on the simplex size.
    args : tuple, optional
        A tuple of arguments to pass to the cost function.
    maxfev : int, optional
        Maximum function evaluations to use. Defaults to 500 times the number
        of parameters.
    simplex_method : str, optional
        Method for generation of initial simplex.

    Returns
    -------
    OptResult
        Object which contains the following attributes:
            xbest (ndarray)   : Optimal values for the optimisation.
            fbest (float)     : Cost function at the optimum.
            niter (int)       : Number of iterations.
            nfev (int)        : Number of function evaluations. Note that in
                                the specific context of NMR optimisation, this
                                is in general not equal to the number of
                                experiments acquired.
            simplex (ndarray) : (N+1, N)-sized matrix of the final simplex.
            fvals (ndarray)   : List of corresponding cost functions at each
                                point of the simplex.
            message (str)     : Message indicating reason for termination.

    Notes
    -----
    The scipy implementation has a few tricks which are useful in making the
    actual computation run faster. However, here we have ignored these in
    favour of readability, since the speed of the optimisation is largely
    limited by the acquisition time of the NMR experiment.
    """

    # Convert x0 to vector
    x0 = np.asfarray(x0).flatten()
    xtol = np.asfarray(xtol).flatten()
    N = x0.size

    # Default maxiter and maxfev. We could make this customisable in future.
    # For example, we could use TopSpin's `expt' to calculate the duration of
    # one experiment, and then set maxiter to not overshoot a given time.
    maxiter = 500 * N
    if maxfev <= 0:
        maxfev = 500 * N
    # Decorate the cost function to raise MaxFevalsReached
    cf = deco_cf(maxfev)(cf)

    # Check length of xtol
    if len(x0) != len(xtol):
        raise ValueError("Nelder-Mead: x0 and xtol have incompatible lengths")

    # Create and initialise simplex object.
    sim = Simplex(x0, method=simplex_method, length=MAGIC_TOL * 10)
    # Number of iterations. Function evaluations are stored as cf.calls.
    niter = 0

    # Set up parameters for Nelder-Mead.
    # Notation follows that used in Section 8.1 of Kelley, 'Iterative Methods
    # for Optimization'.
    mu_ic = -0.5   # Inside contraction parameter
    mu_oc = 0.5    # Outside contraction parameter
    mu_r = 1       # Reflect parameter
    mu_e = 2       # Expansion parameter

    # Helper function.
    def xnew(mu, sim):
        return ((1 + mu) * sim.xbar()) - (mu * sim.xworst())

    def converged(sim, xtol):
        """
        Convergence criteria. To be converged, each dimension of the simplex
        must have a range smaller than or equal to the corresponding xtol in
        that dimension.
        """
        simplex_range = np.amax(sim.x, axis=0) - np.amin(sim.x, axis=0)
        return all(np.less_equal(simplex_range, xtol))
        # Scipy convergence criteria. Assumes that the simplex is already
        # sorted. It is slightly looser (i.e. will converge before mine),
        # but hardly makes a difference to the average fevals (tested on
        # Rosenbrock function).
        #
        # return np.max(np.ravel(np.abs(sim[1:] - sim[0]))) <= xtol[0]

    # Create temporary list of points evaluated during this iteration (and
    # their corresponding cost function values).
    iter_xs, iter_fs = [], []
    try:
        # Evaluate the cost function for the initial simplex.
        # Steps 1 and 2 in Algorithm 8.1.1
        for i in range(N + 1):
            sim.f[i] = cf(sim.x[i], *args)
            # Sort simplex
            sim.sort()

        # Main loop.
        while not converged(sim, xtol):
            # Proceed to next iteration.
            niter += 1
            sim.sort()  # for good measure

            # Check number of iterations.
            if niter >= maxiter:
                raise MaxItersReached

            # Reset both lists.
            iter_xs, iter_fs = [], []

            # Step 3(a)
            x_r = xnew(mu_r, sim)  # shorthand for x(mu_r)
            iter_xs.append(x_r)
            f_r = cf(x_r, *args)
            iter_fs.append(f_r)

            # Step 3(b): Reflect (+ 3g if needed)
            if sim.f[0] <= f_r and f_r < sim.f[N - 1]:
                sim.replace_worst(x_r, f_r)
                sim.sort()  # Step 3(g)
                continue

            # Step 3(c): Expand (+ 3g if needed)
            if f_r < sim.f[0]:
                x_e = xnew(mu_e, sim)
                iter_xs.append(x_e)
                f_e = cf(x_e, *args)
                iter_fs.append(f_e)
                if f_e < f_r:
                    sim.replace_worst(x_e, f_e)
                else:
                    sim.replace_worst(x_r, f_r)
                sim.sort()  # Step 3(g)
                continue

            # Step 3(d): Outside contraction (+ 3f and 3g if needed)
            if sim.f[N - 1] <= f_r and f_r < sim.f[N]:
                x_oc = xnew(mu_oc, sim)
                iter_xs.append(x_oc)
                f_c = cf(x_oc, *args)
                iter_fs.append(f_c)
                if f_c <= f_r:
                    sim.replace_worst(x_oc, f_c)
                    sim.sort()  # Step 3(g)
                    continue
                else:
                    if cf.calls >= maxfev - N:             # Step 3(f)
                        raise MaxFevalsReached
                    sim.shrink()
                    for i in range(1, N + 1):
                        sim.f[i] = cf(sim.x[i], *args)
                    sim.sort()  # Step 3(g)
                    continue

            # Step 3(e): Inside contraction (+ 3f and 3g if needed)
            if f_r >= sim.f[N]:
                x_ic = xnew(mu_ic, sim)
                iter_xs.append(x_ic)
                f_c = cf(x_ic, *args)
                iter_fs.append(f_c)
                if f_c < sim.f[N]:
                    sim.replace_worst(x_ic, f_c)
                    sim.sort()  # Step 3(g)
                    continue
                else:
                    if cf.calls >= maxfev - N:             # Step 3(f)
                        raise MaxFevalsReached
                    sim.shrink()
                    for i in range(1, N + 1):
                        sim.f[i] = cf(sim.x[i], *args)
                    sim.sort()  # Step 3(g)
                    continue
        # END while loop
    except MaxItersReached:
        message = MESSAGE_OPT_MAXITER_REACHED
    except MaxFevalsReached:
        message = MESSAGE_OPT_MAXFEV_REACHED
    except CostFunctionError as e:
        # Check if a value was returned: if so, add it to iter_fs (iter_xs will
        # already have been updated prior to cost function calculation).
        if isinstance(e.cf_val, (int, float)):
            iter_fs.append(e.cf_val)
        message = MESSAGE_OPT_PREMATURE_TERMINATION
        if e.message.strip() != "":
            message += ("\nReason: " + e.message)
    else:
        message = MESSAGE_OPT_SUCCESS

    # sort the simplex in ascending order of fvals
    sim.sort()
    # Check whether the simplex or iter_fs has the lowest cost function.
    if len(iter_fs) != 0 and np.amin(iter_fs) < sim.f[0]:
        xbest, fbest = iter_xs[np.argmin(iter_fs)], np.amin(iter_fs)
    else:
        xbest, fbest = sim.x[0], sim.f[0]

    return OptResult(xbest=xbest, fbest=fbest,
                     niter=niter, nfev=cf.calls,
                     simplex=sim.x, fvals=sim.f,
                     message=message)


def multid_search(cf, x0, xtol, scaled_lb, scaled_ub,
                  args=(), maxfev=0, simplex_method="spendley"):
    """
    Multidimensional search optimiser, as described in Secion 8.2 of Kelley,
    "Iterative Methods for Optimization".

    Parameters
    ----------
    cf : function
        The cost function. For POISE, this means acquire_nmr(), not the
        user-defined cost function. However in general, this can be any cost
        function. The cost function *must* be decorated with deco_count() (for
        POISE, this is already done).
    x0 : ndarray or list
        Initial point for optimisation. This should already be scaled.
    xtol : ndarray or list
        Tolerances for each optimisation dimension. This should already be
        scaled.
    scaled_lb : ndarray
        Scaled lower bounds for the optimisation.
    scaled_ub : ndarray
        Scaled upper bounds for the optimisation. This is used to place an
        upper bound on the simplex size.
    args : tuple, optional
        A tuple of arguments to pass to the cost function.
    maxfev : int, optional
        Maximum function evaluations to use. Defaults to 500 times the number
        of parameters.
    simplex_method : str, optional
        Method for generation of initial simplex.

    Returns
    -------
    OptResult
        Object which contains the following attributes:
            xbest (ndarray)   : Optimal values for the optimisation.
            fbest (float)     : Cost function at the optimum.
            niter (int)       : Number of iterations.
            nfev (int)        : Number of function evaluations. Note that in
                                the specific context of NMR optimisation, this
                                is in general not equal to the number of
                                experiments acquired.
            simplex (ndarray) : (N+1, N)-sized matrix of the final simplex.
            fvals (ndarray)   : List of corresponding cost functions at each
                                point of the simplex.
            message (str)     : Message indicating reason for termination.
    """
    # Convert x0 to vector
    x0 = np.asfarray(x0).flatten()
    xtol = np.asfarray(xtol).flatten()
    N = x0.size

    # Default maxiter and maxfev. We could make this customisable in future.
    # For example, we could use TopSpin's `expt' to calculate the duration of
    # one experiment, and then set maxiter to not overshoot a given time.
    maxiter = 500 * N
    if maxfev <= 0:
        maxfev = 500 * N
    # Decorate cost function so that it handles MaxFevalsReached and
    # CostFunctionError exceptions correctly.
    cf = deco_cf(maxfev)(cf)

    # Check length of xtol
    if len(x0) != len(xtol):
        raise ValueError("Multidimensional search: x0 and xtol have "
                         "incompatible lengths")

    # Create and initialise simplex object.
    sim = Simplex(x0, method=simplex_method, length=MAGIC_TOL * 10)
    # Number of iterations. Function evaluations are stored as cf.calls.
    niter = 0

    # Set up parameters for Nelder-Mead.
    mu_e = 2       # Expansion parameter
    mu_c = 0.5     # Contraction parameter

    def converged(sim, xtol):
        """
        Convergence criteria. To be converged, each dimension of the simplex
        must have a range smaller than or equal to the corresponding xtol in
        that dimension.
        """
        simplex_range = np.amax(sim.x, axis=0) - np.amin(sim.x, axis=0)
        return all(np.less_equal(simplex_range, xtol))
        # Scipy convergence criteria. Assumes that the simplex is already
        # sorted. It is slightly looser (i.e. will converge before mine),
        # but hardly makes a difference to the average fevals (tested on
        # Rosenbrock function).
        #
        # return np.max(np.ravel(np.abs(sim[1:] - sim[0]))) <= xtol[0]

    # Create temporary list of points evaluated during this iteration (and
    # their corresponding cost function values).
    iter_xs, iter_fs = [], []
    # Flag to indicate whether we should break out of the main loop. If a
    # CostFunctionError is raised in the previous iteration, this will be
    # incremented.
    break_on_next_iter = 0

    try:
        # Evaluate the cost function for the initial simplex.
        # Steps 1 and 2 in Algorithm 8.2.1
        for i in range(N + 1):
            sim.f[i], break_increment = cf(sim.x[i], *args)
            # Sort simplex
            sim.sort()

        # Main loop.
        while not converged(sim, xtol):
            niter += 1
            sim.sort()  # for good measure

            # Check number of iterations
            if niter >= maxiter:
                raise MaxItersReached

            # Reset the xs and fs lists.
            iter_xs, iter_fs = [], []

            # Step 3(a): Reflect
            r_j = np.zeros((N, N))
            f_r_j = np.zeros(N)
            for j in range(1, N + 1):
                r_j[j - 1] = sim.x[0] - (sim.x[j] - sim.x[0])
                iter_xs.append(r_j[j - 1])
                f_r_j[j - 1] = cf(r_j[j - 1], *args)
                iter_fs.append(f_r_j[j - 1])

            # Step 3(b): Expand
            if sim.f[0] > np.amin(f_r_j):
                e_j = np.zeros((N, N))
                f_e_j = np.zeros(N)
                for j in range(1, N + 1):
                    e_j[j - 1] = sim.x[0] - mu_e * (sim.x[j] - sim.x[0])
                    iter_xs.append(e_j[j - 1])
                    f_e_j[j - 1] = cf(e_j[j - 1], *args)
                    iter_fs.append(f_e_j[j - 1])
                # Replace the values, 3(b)(ii)
                if np.amin(f_r_j) > np.amin(f_e_j):
                    for j in range(1, N + 1):
                        sim.x[j] = e_j[j - 1]
                        sim.f[j] = f_e_j[j - 1]
                else:
                    for j in range(1, N + 1):
                        sim.x[j] = r_j[j - 1]
                        sim.f[j] = f_r_j[j - 1]
                sim.sort()  # Step 3(d)
                continue
            # Step 3(c): Contract
            else:
                for j in range(1, N + 1):
                    sim.x[j] = sim.x[0] + mu_c * (sim.x[j] - sim.x[0])
                    iter_xs.append(sim.x[j])
                    sim.f[j] = cf(sim.x[j], *args)
                    iter_fs.append(sim.f[j])
                sim.sort()  # Step 3(d)
                continue
    except MaxItersReached:
        message = MESSAGE_OPT_MAXITER_REACHED
    except MaxFevalsReached:
        message = MESSAGE_OPT_MAXFEV_REACHED
    except CostFunctionError as e:
        # Check if a value was returned: if so, add it to iter_fs (iter_xs will
        # already have been updated prior to cost function calculation).
        if isinstance(e.cf_val, (int, float)):
            iter_fs.append(e.cf_val)
        message = MESSAGE_OPT_PREMATURE_TERMINATION
        if e.message.strip() != "":
            message += ("\nReason: " + e.message)
    else:
        message = MESSAGE_OPT_SUCCESS

    # sort the simplex in ascending order of fvals
    sim.sort()
    # Check whether the simplex or iter_fs has the lowest cost function.
    if len(iter_fs) != 0 and np.amin(iter_fs) < sim.f[0]:
        xbest, fbest = iter_xs[np.argmin(iter_fs)], np.amin(iter_fs)
    else:
        xbest, fbest = sim.x[0], sim.f[0]

    return OptResult(xbest=xbest, fbest=fbest,
                     niter=niter, nfev=cf.calls,
                     simplex=sim.x, fvals=sim.f,
                     message=message)


def pybobyqa_interface(cf, x0, xtol, scaled_lb, scaled_ub,
                       args=(), maxfev=0):
    """
    Interface to pybobyqa.solve() which takes similar arguments to the other
    two optimisation functions and returns an OptResult object.

    Parameters
    ----------
    cf : function
        The cost function. For POISE, this means acquire_nmr(), not the
        user-defined cost function. However in general, this can be any cost
        function.
    x0 : ndarray or list
        Initial point for optimisation. This must already be scaled by tols.
    xtol : ndarray or list
        Tolerances for each optimisation dimension. This is actually not used
        at all and is only here so that there is a unified interface for all
        three optimisers.
    args : tuple, optional
        A tuple of arguments to pass to the cost function.
    simplex_method : str, optional
        Method for generation of initial simplex.
    scaled_lb : ndarray, optional
        Scaled lower bounds for the optimisation.
    scaled_ub : ndarray, optional
        Scaled upper bounds for the optimisation. This is used to place an
        upper bound on the simplex size.

    Returns
    -------
    OptResult
        Object which contains the following attributes:
            xbest (ndarray)   : Optimal values for the optimisation.
            fbest (float)     : Cost function at the optimum.
            niter (int)       : Number of iterations. solve() does not actually
                                provide this information directly, so this
                                value is simply set to zero. (It is indirectly
                                available if diagnostic info is requested.)
            nfev (int)        : Number of function evaluations. Note that in
                                the specific context of NMR optimisation, this
                                is in general not equal to the number of
                                experiments acquired.
            message (str)     : Message indicating reason for termination.
    """
    x0 = np.asfarray(x0).flatten()
    # Calculate maxfev
    N = x0.size
    if maxfev <= 0:
        maxfev = 500 * N
    # Run the optimisation, using PyBOBYQA's bounds keyword arguments.
    bounds = (scaled_lb, scaled_ub)
    min_ub = np.min(scaled_ub)
    try:
        pb_sol = pb.solve(cf, x0, args=args,
                          rhobeg=min(MAGIC_TOL * 10, min_ub * 0.499),
                          rhoend=MAGIC_TOL,
                          maxfun=maxfev,
                          bounds=bounds, objfun_has_noise=True,
                          user_params={'restarts.use_restarts': False})
    except CostFunctionError as e:
        # Catch user-thrown errors. Because the BOBYQA optimiser is opaque,
        # there's no way we can return any useful info to the user, so we just
        # convert it into a RuntimeError which will be passed on to the
        # frontend.
        raise RuntimeError(e.message) from None
    # Catch Py-BOBYQA complaints if the optimiser exited with failure.
    if pb_sol.flag in [pb_sol.EXIT_INPUT_ERROR,
                       pb_sol.EXIT_TR_INCREASE_ERROR,
                       pb_sol.EXIT_LINALG_ERROR]:
        raise RuntimeError(pb_sol.msg)
    # If we reached here, it means the optimisation completed successfully.
    # We just need to coerce the returned information into our OptResult
    # format, so that the backend sees a unified interface for all optimisers.
    # Note that niter is not applicable to PyBOBYQA.
    # TODO: Convert PyBOBYQA's messages (success, maxiter, maxfev) back into
    # our standardised messages, so that the frontend can check what happened.
    if pb_sol.flag == pb_sol.EXIT_SUCCESS:
        msg = MESSAGE_OPT_SUCCESS
    elif pb_sol.flag == pb_sol.EXIT_MAXFUN_WARNING:
        msg = MESSAGE_OPT_MAXFEV_REACHED
    elif pb_sol.flag == pb_sol.EXIT_SLOW_WARNING:
        msg = "Optimisation terminated (slow progress)."
    elif pb_sol.flag == pb_sol.EXIT_FALSE_SUCCESS_WARNING:
        msg = "Optimisation terminated (max false good steps)."
    else:
        msg = pb_sol.msg
    return OptResult(xbest=pb_sol.x, fbest=pb_sol.f,
                     niter=0, nfev=pb_sol.nf,
                     message=msg)



"""
poisecalpy.py
-------------

Download from: https://git.io/JUHOU

Python script which automates p1 optimisation using POISE. Behaves similarly to
pulsecal: optimises the value of p1 in expno 99999, then sets the optimised
value to the current dataset and shows a message.

This does not have the command-line argument functionality of the poisecal AU
programme, and tends to suffer from some edge-case bugs which are not present
in poisecal. This is generally a result of TopSpin's Python programmes being
more buggy than AU programmes. Nonetheless, as a result of this, we recommend
using the poisecal AU programme instead of this (which is retained mainly to
serve as a illustrative example in the documentation).

SPDX-License-Identifier: GPL-3.0-or-later
"""

# Read in F1P and F2P from current dataset.
f1p = GETPAR("F1P")
f2p = GETPAR("F2P")
# Use EXPNO 99999 in the current folder for optimisation.
old_dataset = CURDATA()
opt_dataset = CURDATA()
opt_dataset[1] = "99999"
# Create new dataset and move to it.
NEWDATASET(opt_dataset, None, "PROTON")
RE(opt_dataset)
# Set some key parameters. Notice that these lines can be cut if an appropriate
# parameter set is set up beforehand (and loaded using NEWDATASET()).
XCMD("getprosol")
PUTPAR("PULPROG", "zg")
PUTPAR("NS", "1")
PUTPAR("DS", "0")
PUTPAR("D 1", "1")
PUTPAR("RG", "1")
XCMD("s f1p {}".format(f1p))  # PUTPAR("status F1P") doesn't work.
XCMD("s f2p {}".format(f2p))  # Even though the documentation says it should.
# Run optimisation.
XCMD("poise p1cal -a bobyqa -q")
# POISE stores the optimised value in p1 after it's done. We can retrieve it
# here. Don't try to get the *status* parameter, since that is not the
# optimised value (it is the value used for the last function evaluation!)
p1opt = float(GETPAR("P 1"))/4
# Move back to old dataset and set p1 to optimised value.
RE(old_dataset)
PUTPAR("P 1", str(p1opt))
ERRMSG("Optimised value of p1: {:.3f}".format(p1opt))
# A TopSpin quirk: using MSG() will block subsequent commands until user hits
# "OK". You can use ERRMSG(), as is done here. There are other ways around
# this, see MSG_nonmodal() in the POISE frontend script.
# (Optional) Run acquisition.
# ZG()



"""
backend.py
----------

Backend POISE script which carries out the optimisation, passing values to the
frontend to trigger acquisition.

SPDX-License-Identifier: GPL-3.0-or-later
"""

import os
import sys
import json
from traceback import print_exc
from datetime import datetime
from pathlib import Path
from collections import namedtuple
from contextlib import contextmanager

import numpy as np

# Enable relative imports when invoked directly as __main__ by TopSpin.
# cf. PEP 366 and https://stackoverflow.com/a/54490918
# This allows us to import the optpoise.py that is inside $ts/py/user, instead
# of the one inside site-packages. Essentially, that allows us to maintain
# internal consistency as the version of code that is run is always the one
# inside $ts/py/user.
if __name__ == "__main__" and __package__ is None:
    __package__ = "poise_backend"
    sys.path.insert(1, str(Path(__file__).parents[1].resolve()))
    __import__(__package__)

from .optpoise import (scale, unscale, deco_count,
                       nelder_mead, multid_search, pybobyqa_interface)
from .shared import _g
from .cfhelpers import *
from . import costfunctions
from . import costfunctions_user

Routine = namedtuple("Routine", "name pars lb ub init tol cf au")


@contextmanager
def pidfile():
    """
    Context manager that creates a '.pid<PID>' file inside the poise_backend
    directory. Deletes the file once the backend exits.
    """
    pid = os.getpid()
    pid_fname = Path(__file__).resolve().expanduser().parent / f".pid{pid}"
    pid_fname.touch()
    try:
        yield
    finally:
        if pid_fname.exists():
            pid_fname.unlink()


def main_wrapper():
    """
    Wrapper for main(). This performs several tasks:
      1. Reads in the global variables printed by the frontend.
      2. Implements a context manager to print the backend PID to a file
         ($ts/py/user/poise_backend/pid.log), deleting the file after the
         backend completes.
      3. Catches all exceptions, propagates them to the frontend by printing to
         stdout, and prints the full traceback to the backend error log.
    """
    with pidfile() as _:
        try:
            from .shared import _g
            # Set global variables by reading in input from frontend.
            _g.optimiser = input()
            _g.routine_id = input()
            _g.p_spectrum = Path(input())
            _g.maxfev = int(input())
            _g.p_optlog = _g.p_spectrum.parents[1] / "poise.log"
            _g.p_errlog = _g.p_spectrum.parents[1] / "poise_err_backend.log"
            # Run main routine.
            main()
        except Exception as e:
            # Because the frontend is only reading one line at a time, there's
            # no point in printing the entire traceback. Thus we just print a
            # very short summary line.
            print(f"Backend exception: {type(e).__name__}({repr(e.args)})")
            # Then print it to the errlog.
            with open(_g.p_errlog, "a") as ferr:
                print("======= From backend =======", file=ferr)
                print(datetime.now().strftime("%Y-%m-%d %H:%M:%S"), file=ferr)
                print_exc(file=ferr)
                print("\n\n", file=ferr)


def main():
    """
    Main routine.
    """
    tic = datetime.now()

    # Load the routine and cost function.
    routine, cost_function = get_routine_cf(_g.routine_id)

    # Choose the optimisation function. optpoise implements a PyBOBYQA
    # interface so that the returned result has the same attributes as our
    # other optimisers.
    optimfndict = {"nm": nelder_mead,
                   "mds": multid_search,
                   "bobyqa": pybobyqa_interface
                   }
    try:
        optimfn = optimfndict[_g.optimiser.lower()]
    except KeyError:
        raise ValueError(f"Invalid optimiser {_g.optimiser} specified.")

    # Scale the initial values and tolerances
    npars = len(routine.pars)
    scaled_x0, scaled_lb, scaled_ub, scaled_xtol = scale(routine.init,
                                                         routine.lb,
                                                         routine.ub,
                                                         routine.tol,
                                                         scaleby="tols")

    # Some logging
    with open(_g.p_optlog, "a") as log:
        print("\n\n\n", file=log)
        print("=" * 40, file=log)
        print(datetime.now().strftime("%Y-%m-%d %H:%M:%S"), file=log)
        fmt = "{:25s} - {}"
        print(fmt.format("Routine name", routine.name), file=log)
        print(fmt.format("Optimisation parameters", routine.pars), file=log)
        print(fmt.format("Cost function", routine.cf), file=log)
        print(fmt.format("AU programme", routine.au), file=log)
        print(fmt.format("Initial values", routine.init), file=log)
        print(fmt.format("Lower bounds", routine.lb), file=log)
        print(fmt.format("Upper bounds", routine.ub), file=log)
        print(fmt.format("Tolerances", routine.tol), file=log)
        print(fmt.format("Optimisation algorithm", _g.optimiser), file=log)
        print("", file=log)
        fmt = "{:^10s}  " * (npars + 1)
        print(fmt.format(*routine.pars, "cf"), file=log)
        print("-" * 12 * (npars + 1), file=log)

    # Get F1P/F2P parameters
    x, y = getpar("F1P"), getpar("F2P")
    if isinstance(x, float) and isinstance(y, float):  # 1D
        if x != 0 and y != 0:
            _g.spec_f1p, _g.spec_f2p = x, y
    elif isinstance(x, np.ndarray) and isinstance(y, np.ndarray):  # 2D
        if not np.array_equal(x, [0, 0]) and not np.array_equal(y, [0, 0]):
            _g.spec_f1p, _g.spec_f2p = x, y

    # Set up optimisation arguments
    optimargs = (cost_function, routine)
    # Carry out the optimisation
    opt_result = optimfn(acquire_nmr, scaled_x0, scaled_xtol,
                         scaled_lb, scaled_ub,
                         args=optimargs, maxfev=_g.maxfev)

    # Tell frontend script that the optimisation is done
    best_values = unscale(opt_result.xbest, routine.lb,
                          routine.ub, routine.tol, scaleby="tols")
    print(f"optima: {' '.join([str(i) for i in best_values])}")
    # Strip newlines from the opt result message, just in case (because the
    # frontend only expects one line of text here, feeding it more than one
    # line of text will confuse it)
    print(opt_result.message.replace("\n", " ").replace("\r", " "))

    # More logging
    toc = datetime.now()
    time_taken = str(toc - tic).split(".")[0]  # remove microseconds
    with open(_g.p_optlog, "a") as log:
        print("", file=log)
        fmt = "{:27s} - {}"
        print(fmt.format("Best values found", best_values.tolist()), file=log)
        print(fmt.format("Cost function at minimum", opt_result.fbest),
              file=log)
        print(fmt.format("Number of spectra ran", acquire_nmr.calls),
              file=log)
        print(fmt.format("Total time taken", time_taken), file=log)


def get_routine_cf(routine_id, p_routine_dir=None):
    """
    First finds the routine file and instantiates a Routine. Then finds the
    associated cost function script and exec's it, which defines
    cost_function().

    Parameters
    ----------
    routine_id : str
        Name of the routine being used.

    Returns
    -------
    Routine
        The Routine object requested.
    function
        The cost function object associated with the routine.

    Other Parameters
    ----------------
    p_routine_dir : pathlib.Path, optional
        Path to the folder containing the routines. Defaults to the
        poise_backend/routines directory. This parameter only exists for
        unit test purposes and should not otherwise be used.
    """
    # Load the routine.
    p_routine_dir = p_routine_dir or (_g.p_poise / "routines")
    with open(p_routine_dir / (routine_id + ".json"), "rb") as f:
        routine = Routine(**json.load(f))

    # Load the cost function. Try to get from user first.
    cost_function = getattr(costfunctions_user, routine.cf, None)
    # If it failed, get from system.
    if cost_function is None:
        cost_function = getattr(costfunctions, routine.cf, None)
    # If that failed too, then error out.
    if cost_function is None:
        raise AttributeError(f"No such cost function {routine.cf}.")
    return routine, cost_function


@deco_count
def acquire_nmr(x, cost_function, routine):
    """
    This is the function which is actually passed to the optimisation function
    as the "cost function", and is responsible for triggering acquisition in
    the frontend.

    Briefly, this function does the following:

     - Returns np.inf immediately if the values are outside the given bounds.
     - Otherwise, prints the values to stdout, which triggers acquisition by
       the frontend.
     - Waits for the frontend to pass the message "done" back, indicating that
       the spectrum has been acquired and processed.
     - Calculates the cost function using the user-defined cost_function().
     - Performs logging throughout.

    Parameters
    ----------
    x : ndarray
        Scaled values to be used for spectrum acquisition and cost function
        evaluation.
    cost_function : function
        User-defined cost function object.
    routine : Routine
        The active optimisation routine.

    Returns
    -------
    cf_val : float
        Value of the cost function.
    """
    unscaled_val = unscale(x, routine.lb, routine.ub,
                           routine.tol, scaleby="tols")
    # Format string for logging.
    fstr = "{:^10.4f}  " * (len(x) + 1)

    with open(_g.p_optlog, "a") as logf:
        # Enforce constraints on optimisation.
        # This doesn't need to be done for BOBYQA, because we pass the `bounds`
        # parameter, which automatically stops it from sampling outside the
        # bounds. If we *do* enforce the constraints on BOBYQA, this can lead
        # to bad errors, as due to floating-point inaccuracy sometimes it tries
        # to sample a point that is ___just___ outside of the bounds (see #39).
        # Instead, we should just let it evaluate the point as usual.
        if (_g.optimiser in ["nm", "mds"] and
                (np.any(unscaled_val < routine.lb)
                 or np.any(unscaled_val > routine.ub))):
            # Set the value of the cost function to infinity.
            cf_val = np.inf
            # Log that.
            print(fstr.format(*unscaled_val, cf_val), file=logf)
            # Return immediately.
            return cf_val

        # Print unscaled values, prompting frontend script to start acquisition
        print("values: " + " ".join([str(i) for i in unscaled_val]))
        # Wait for acquisition to complete, then calculate cost function
        signal = input()  # frontend prints "done" here
        # Set p_spectrum according to which spectrum the frontend evaluated.
        # This is important when using the separate_expnos option.
        _g.p_spectrum = Path(input())  # frontend prints path to active spec.
        # Evaluate the cost function, log, pass the cost function value back
        # to the frontend (it's stored in the `TI` parameter), and return.
        if signal == "done":
            _g.nfev += 1
            fstr = "{:^10.4f}  " * (len(x) + 1)
            try:
                cf_val = cost_function()
            except CostFunctionError as e:
                # Log the point and cost function
                if isinstance(e.cf_val, (int, float)):
                    print(fstr.format(*unscaled_val, e.cf_val), file=logf)
                else:
                    fstr2 = "{:^10.4f}  " * len(x)
                    print(fstr2.format(*unscaled_val), file=logf)
                # Log the exception and return control to optimiser.
                print(e.message, file=logf)
                raise
            else:
                print(fstr.format(*unscaled_val, cf_val), file=logf)
                print(f"cf: {cf_val}")  # send back to frontend
                return cf_val    # return control to optimiser
        else:
            # This really shouldn't happen.
            raise ValueError(f"Invalid signal passed from frontend: {signal}")


if __name__ == "__main__":
    main_wrapper()



"""
SPDX-License-Identifier: GPL-3.0-or-later
"""

import json
import re
from pathlib import Path

import numpy as np
import pandas as pd


def parse_log(fname="."):
    """
    Parse a poise.log file.

    Parameters
    ----------
    fname : (optional) str or pathlib.Path or int
        Path to poise.log file, or the folder containing it (this would be the
        TopSpin EXPNO folder). If not specified assumes the current working
        directory. If passed as an int, assumes "./<fname>" (i.e. expno X in
        the current working directory).

    Returns
    -------
    log_df : pandas.DataFrame
        Dataframe with rows corresponding to optimisations which successfully
        terminated. The time taken is given in seconds.
    """
    # Handle the int case
    if isinstance(fname, int):
        fname = f"./{fname}"
    # Check if the file exists
    fname = Path(fname).resolve()
    if fname.is_dir():
        fname = fname / "poise.log"
    if not fname.exists():
        raise FileNotFoundError(f"The log file '{fname}' was not found.")

    empty_run = {"routine": None, "init": None, "cf": None, "algo": None,
                 "au": None, "lb": None, "ub": None, "tol": None,
                 "xbest": None, "feval": None, "elapsed": None,
                 }
    current_run = dict(empty_run)  # make a copy
    all_runs = []   # List of all successful optimisation runs

    def parse_list_params(line):
        """
        Extracts init, lb, ub, tol from the line.
        """
        # Get the second half of the line
        rhs = line.split("-", maxsplit=1)[1].strip()
        # Convert '[4.]' to '[4]', otherwise json.loads() trips up
        rhs = re.sub(r"\.([^\d])", r"\1", rhs)
        # Evaluate the string and remove the list if it's a singleton
        rhs = json.loads(rhs)
        return rhs[0] if len(rhs) == 1 else rhs

    with open(fname, "r") as fp:
        for line in fp:
            # Check if there is a completed optimisation stored.
            # If so, add it to all_runs then reset the current run.
            if current_run["elapsed"] is not None:
                all_runs.append(current_run)
                current_run = dict(empty_run)
            # Start of a new optimisation, reset all values
            if "===================" in line:
                current_run = dict(empty_run)
            # Parse lines
            elif line.startswith("Routine name"):
                current_run["routine"] = line.split("-", maxsplit=1)[1].strip()
            elif line.startswith("Optimisation parameters"):
                p = line.split("-", maxsplit=1)[1].strip().replace("'", '"')
                current_run["param"] = json.loads(p)
            elif line.startswith("Initial values"):
                current_run["init"] = parse_list_params(line)
            elif line.startswith("Lower bounds"):
                current_run["lb"] = parse_list_params(line)
            elif line.startswith("Upper bounds"):
                current_run["ub"] = parse_list_params(line)
            elif line.startswith("Tolerances"):
                current_run["tol"] = parse_list_params(line)
            elif line.startswith("Cost function  "):
                current_run["cf"] = line.split("-", maxsplit=1)[1].strip()
            elif line.startswith("AU programme"):
                current_run["au"] = line.split("-", maxsplit=1)[1].strip()
            elif line.startswith("Optimisation algorithm"):
                current_run["algo"] = line.split("-", maxsplit=1)[1].strip()
            elif line.startswith("Best values found"):
                current_run["xbest"] = parse_list_params(line)
            elif line.startswith("Cost function at"):
                fbest = line.split("-", maxsplit=1)[1].strip()
                current_run["fbest"] = float(fbest)
            elif line.startswith("Number of spectra ran"):
                feval = line.split("-", maxsplit=1)[1].strip()
                current_run["feval"] = int(feval)
            elif line.startswith("Total time taken"):
                elapsed = line.split("-", maxsplit=1)[1].strip()
                elapsed = [int(i) for i in elapsed.split(":")]
                current_run["elapsed"] = (3600*elapsed[0]
                                          + 60*elapsed[1]
                                          + elapsed[2])

    # Check one more time at the end of the file to make sure that the
    # last entry is recorded.
    if current_run["elapsed"] is not None:
        all_runs.append(current_run)
        current_run = dict(empty_run)

    # Construct dataframe and return.
    data = {}
    data["routine"] = [run["routine"] for run in all_runs]
    data["initial"] = [run["init"] for run in all_runs]
    data["param"] = [run["param"] for run in all_runs]
    data["lb"] = [run["lb"] for run in all_runs]
    data["ub"] = [run["ub"] for run in all_runs]
    data["tol"] = [run["tol"] for run in all_runs]
    data["algorithm"] = [run["algo"] for run in all_runs]
    data["costfn"] = [run["cf"] for run in all_runs]
    data["auprog"] = [run["au"] for run in all_runs]
    data["optimum"] = [run["xbest"] for run in all_runs]
    data["fbest"] = [run["fbest"] for run in all_runs]
    data["nfev"] = [run["feval"] for run in all_runs]
    data["time"] = [run["elapsed"] for run in all_runs]
    return pd.DataFrame(data=data)


def xmean(series):
    """
    An aggregator function that allows you to average lists in DataFrame
    columns. This is useful for analysing optimisation runs with multiple
    parameters.

    For example:

    >>> df = parse_log(logfile)
    >>> # works for 1D, but not for nD optimisations
    >>> df.groupby("algorithm").mean()
    >>> # works for all dimensions.
    >>> df.groupby("algorithm").agg(xmean)
    """
    return np.mean(np.array(series.to_list()), axis=0).tolist()



"""
dosy_opt.py
-----------

Download from: https://git.io/JUHOY

Script that optimises parameters for a diffusion experiment. Currently works
with oneshot DOSY (Pelta et al., Magn. Reson. Chem. 2002, 40 (13), S147-S152,
DOI: 10.1002/mrc.1107), as well as the stegp1s and ledgp2s DOSY variants which
are standard TopSpin pulse programmes.

This is meant to be called from the TopSpin command line on a 2D diffusion
experiment.

Note that this script does not tweak the gradient lengths ("little delta"), so
these should be set to a sensible value (e.g. 1 ms) before running this script.

The algorithm is as follows:

 1. Create 1D versions of the diffusion experiment, one with the minimum
    gradient (10%, the "reference spectrum", expno 99998) and one with the
    maximum gradient (80%, "optimisation spectrum", expno 99999).
 2. Acquires both spectra.
 3. Check whether the opt. spectrum has at least 75% signal attenuation
    relative to the ref. spectrum. If not, increases the diffusion delay
    ("big Delta") and repeats steps 1 and 2.
 4. If the attenuation is sufficient, it then runs a POISE optimisation on the
    opt. spectrum to find the gradient strength at which 75% signal attenuation
    is achieved.
 5. Once this gradient strength has been found, returns to the 2D experiment
    and runs the `dosy` Bruker AU programme using the new parameter.

SPDX-License-Identifier: GPL-3.0-or-later
"""

# Dictionary of DOSY pulse programmes, to be entered in the form
#    "2d_pulprog": ["1d_pulprog", "GRAD_AMPLITUDE_PARAM", "DIFF_DELAY_PARAM"]
# In more detail:
#  - 2d_pulprog is the name of the 2D DOSY pulse sequence.
#  - 1d_pulprog is the name of the corresponding 1D DOSY pulse sequence, which
#     should be just a single increment of the 2D sequence, but with Difframp
#     removed. See the doneshot_nd_jy pulse programmes for an example.
#  - GRAD_AMPLITUDE_PARAM is the parameter name in the 2D DOSY pulse sequence
#     that is multiplied by Difframp. In the 1D DOSY pulse sequence it should
#     not be multiplied by Difframp, so simply corresponds to the gradient
#     strength. Note that you need a space between GPZ and the number.
#  - DIFF_DELAY_PARAM is the parameter name in both 1D and 2D sequences that
#     corresponds to the diffusion delay. Again, a space between D and the
#     number is required.
# Other DOSY variants can be entered here to "register" them and make them
# available for use with this script.
pulprog_dict = {
    # doneshot_nd_jy can be downloaded from https://git.io/JUHOX and
    # https://git.io/JUHOP.
    "doneshot_2d_jy": ["doneshot_1d_jy", "GPZ 1", "D 20"],
    # the following two entries are Bruker standard sequences and already come
    # with TopSpin.
    "stegp1s": ["stegp1s1d", "GPZ 6", "D 20"],
    "ledgp2s": ["ledgp2s1d", "GPZ 6", "D 20"],
}


def main():
    # Look up the 1D pulse programme and associated parameters based on the
    # current 2D pulse programme.
    pp2d = GETPAR("PULPROG")
    if pp2d in pulprog_dict:
        pp1d, gpzparam, dparam = pulprog_dict[pp2d]
    else:
        MSG("dosy_opt: unsupported pulse programme. New pulse programmes can"
            " be added in the dosy_opt.py script.")
        EXIT()

    # Create new routines. We need to do this within this script since the
    # routine depends on the GPZ parameter being optimised; we can't hardcode
    # that in a predefined routine.
    gpzparam_short = gpzparam.replace(" ", "").lower()
    XCMD("poise --create dosy {} 20 80 50 2"
         " dosy poise_1d".format(gpzparam_short))
    XCMD("poise --create dosy_aux {} 79 81 80 0.1"
         " dosy_aux poise_1d".format(gpzparam_short))

    # Get the expno of the 2D dataset, and store the parameter set. This is
    # important because we want the optimisation to inherit most parameters,
    # e.g.  D1, SW, etc. from the parent experiment (which has presumably been
    # set up by the user).
    original_2d_expno = int(CURDATA()[1])
    original_2d_dataset = CURDATA()
    XCMD("wpar dosytemp all")

    # Set up the reference experiment.
    reference_expno = 99998
    reference_dataset = list(original_2d_dataset)
    reference_dataset[1] = str(reference_expno)
    NEWDATASET(reference_dataset, None, "dosytemp")
    RE(reference_dataset)
    PUTPAR("PULPROG", pp1d)
    PUTPAR("PARMODE", "1D")
    PUTPAR("PPARMOD", "1D")
    PUTPAR(gpzparam, "10")
    XCMD("sendgui browse_update_tree")
    # Acquire reference dataset.
    msg_nonmodal("dosy_opt: optimising diffusion delay...")
    XCMD("poise_1d", wait=WAIT_TILL_DONE)

    # Set up optimisation expno.
    optimisation_expno = 99999
    optimisation_dataset = list(original_2d_dataset)   # make a copy
    optimisation_dataset[1] = str(optimisation_expno)
    NEWDATASET(optimisation_dataset, None, "dosytemp")
    RE(optimisation_dataset)
    PUTPAR("PULPROG", pp1d)
    PUTPAR("PARMODE", "1D")
    PUTPAR("PPARMOD", "1D")
    PUTPAR(gpzparam, "80")
    XCMD("sendgui browse_update_tree")

    # Optimise the diffusion delay first.
    while True:
        # Here we use the `--maxfev 1` trick to evaluate a cost function at 80%
        # gradient. The cost function is stored as the `TI` parameter after the
        # experiment has been recorded.
        RE(optimisation_dataset)
        PUTPAR("TI", " ")  # has to be one space, not an empty string (try it)
        XCMD("poise dosy_aux -a nm --maxfev 1 -q", wait=WAIT_TILL_DONE)
        # If TI isn't a valid float, something went wrong with POISE.
        try:
            cf = float(GETPAR("TI"))
        except ValueError:
            EXIT()
        # The dosy_aux cost function is (opt. intensity/ref. intensity) minus
        # 0.25. If the opt. spectrum has been attenuated too little, then this
        # value will be positive.
        if cf > 0:
            # Increase delay, i.e. big Delta, by 100 ms
            current_delay = float(GETPAR(dparam))
            new_delay = current_delay + 0.1
            PUTPAR(dparam, str(new_delay))
            # Set it in the reference spectrum too
            RE(reference_dataset)
            PUTPAR(dparam, str(new_delay))
            # Re-acquire the reference spectrum.
            XCMD("xau poise_1d", wait=WAIT_TILL_DONE)
        # If the cost function is negative, then that means it's sufficiently
        # attenuated, and we can go on to the actual optimisation.
        else:
            break

    # At this point, we know that 80% gradient strength provides at least 75%
    # attenuation. So we can run an optimisation to find the actual value that
    # does provide 75% attenuation (which is somewhere between 10% and 80%).
    RE(optimisation_dataset)
    best_delay = GETPAR(dparam)
    msg_nonmodal("dosy_opt: diffusion delay set to {} seconds.\n"
                 "    now optimising gradient amplitude...".format(best_delay))
    PUTPAR("TI", " ")
    XCMD("poise dosy -a bobyqa -q", wait=WAIT_TILL_DONE)
    # Again, check to make sure that TI is a valid float.
    try:
        float(GETPAR("TI"))
    except ValueError:
        EXIT()
    best_gpz = GETPAR(gpzparam)

    # Jump back to original 2D expt. Set dparam, but not gpzparam as that
    # should always be 100% (the actual gradient amplitude is controlled by
    # Difframp, which will be generated by the `dosy` AU programme).
    RE(original_2d_dataset)
    PUTPAR(dparam, best_delay)
    PUTPAR(gpzparam, "100")
    # Generate Difframp, etc. and then run the AU programme.
    msg_nonmodal("dosy_opt: maximum gradient set to {}%.\n"
                 "    starting 2D experiment with optimised"
                 " parameters...".format(best_gpz))
    XCMD("xau dosy 10 {} 16 l y".format(best_gpz))


def msg_nonmodal(s):
    """
    Show a message box that does not force the user to click OK before the
    programme continues. See also the same function in poise.py (the frontend
    script).
    """
    dialog = mfw.BInfo()
    dialog.setMessage(s)
    dialog.setTitle("dosy_opt")
    dialog.setBlocking(0)  # in ordinary MSG() this is 1
    dialog.show()


if __name__ == "__main__":
    main()



"""
poise.py
--------

Parameter Optimisation by Iterative Spectral Evaluation

Frontend Python script that runs in TopSpin. Type `poise -h` into the TopSpin
command line for help, or visit `https://foroozandehgroup.github.io/nmrpoise`
for a more thorough introduction.

SPDX-License-Identifier: GPL-3.0-or-later
"""

from __future__ import division, with_statement, print_function

import os
import re
import json
import subprocess
import argparse
from datetime import datetime
from traceback import print_exc
from collections import namedtuple

from de.bruker.nmr.mfw.root.UtilPath import getTopspinHome
from de.bruker.nmr.prsc.dbxml.ParfileLocator import getParfileDirs

tshome = getTopspinHome()
p_poise = os.path.join(tshome, "exp/stan/nmr/py/user/poise_backend")
p_backend = os.path.join(p_poise, "backend.py")
p_routines = os.path.join(p_poise, "routines")
p_python3 = r"/usr/local/bin/python"
Routine = namedtuple("Routine", "name pars lb ub init tol cf au")


def main(args):
    """
    Main routine.

    Parameters
    ---------
    args (argparse.Namespace)
        Namespace object returned by parser.parse_args().

    Returns
    -------
    None
    """
    # Check that the folders are valid
    for folder in [p_poise, p_routines]:
        if not os.path.isdir(folder):
            os.makedirs(folder)
    # Check that cost functions folder actually has some cost functions
    _cfs = detect_costfunctions()
    if len(_cfs) == 0:
        err_exit("No cost functions were found. Please define a cost function "
                 ", or reinstall poise to obtain the defaults.")

    # Select an optimisation routine.
    saved_routines = list_files(p_routines, ext=".json")
    # If routine was specified on command-line, check that there actually is a
    # routine with that name.
    if args.routine is not None:
        if args.routine in saved_routines:
            routine_id = args.routine
            routine = load_routine(routine_id)
            prompt_run = False
        else:
            err_exit("The routine '{}' was not found. Use 'poise --list' to "
                     "see all available routines.".format(args.routine))
    # If the routine wasn't specified on command-line, prompt the user for one
    else:
        routine_id = get_routine_id()

    # If get_routine_id() returns None, then a new routine was requested.
    if routine_id is None:
        # Prompt the user later whether they want to run it after creating it.
        prompt_run = True
        # Get the Routine object.
        routine = get_new_routine()
        # Store the routine for future usage
        with open(os.path.join(p_routines, routine.name + ".json"), "wb") as f:
            json.dump(routine._asdict(), f)
        routine_id = routine.name
    # Otherwise, load a saved routine.
    else:
        routine = load_routine(routine_id)
        prompt_run = False

    # Prompt the user whether they want to run it now, but only if they created
    # a new routine from scratch. If they entered an existing routine (either
    # via command-line, or in a dialog box), then we assume they want to run it
    # now.
    if prompt_run and not run_now():
        EXIT()

    # If they do want to run it, then we can carry on.
    # Make sure user has opened a dataset
    if CURDATA() is None:
        err_exit("Please select a dataset!")

    # Keep track of the currently active dataset. We want to make sure that
    # it is _the_ active dataset every time the AU programme starts, or
    # else we risk running `zg` on a different dataset --> overwriting
    # other data!!
    current_dataset = CURDATA()

    # Construct the path to the folder of the current spectrum.
    p_spectrum = make_p_spectrum()
    # Generate the path to the log file. The log file belongs to the folder
    # containing the first expno.
    p_optlog = os.path.normpath(os.path.join(current_dataset[3],
                                             current_dataset[0],
                                             current_dataset[1],
                                             "poise.log"))

    # Issue a warning if dataset is > 2D
    if GETACQUDIM() > 2:
        MSG("Warning: poise is only designed for 1D and 2D data.\n"
            "If you are sure you want to use it on this {}D dataset,"
            " please press OK to continue.".format(GETACQUDIM()))

    # Check that the Routine object itself is valid (i.e. has all the correct
    # keys).
    check_routine(routine)
    # Check that the cost function exists
    saved_cfs = detect_costfunctions().keys()
    if routine.cf not in saved_cfs:
        err_exit("The associated cost function {} could not be"
                 " found.".format(routine.cf))
    # If the AU programme is blank, set it to poise_nd.
    if routine.au == "":
        dim = GETACQUDIM()
        if dim == 1:
            # Namedtuples are immutable, so we need the _replace method.
            routine = routine._replace(au="poise_1d")
        elif dim == 2:
            routine = routine._replace(au="poise_2d")
        else:
            err_exit("You must specify an AU programme when using POISE with"
                     " >2D experiments.")
    # Check that the AU programme exists.
    if not au_exists(routine.au):
        err_exit("The AU programme {} was not found.".format(routine.au))

    # Make sure that Python 3 executable can be found
    check_python3path()

    # Check that the backend script is intact
    if not os.path.isfile(p_backend):
        err_exit("Backend script not found. Please reinstall poise.")

    # Make sure that args.algorithm is a valid algorithm.
    if args.algorithm not in ["nm", "mds", "bobyqa"]:
        # Have to use ERRMSG because MSG() is modal
        ERRMSG("Optimisation algorithm '{}' not found; "
               "using Nelder-Mead instead".format(args.algorithm))
        args.algorithm = "nm"

    # Before we start the main loop, we need to kill off any other backend
    # processes that are still alive. (We don't do it too early, otherwise we
    # could accidentally terminate a POISE run by running something innocuous
    # like poise -l.)
    kill_remaining_backends()

    # We need to catch java.lang.Error throughout the main loop so that cleanup
    # can be performed if the script is killed from within TopSpin. See #23.
    try:
        backend = subprocess.Popen([p_python3, '-u', p_backend],
                                   stdin=subprocess.PIPE,
                                   stdout=subprocess.PIPE)
        # Pass key information to the backend script
        print(args.algorithm, file=backend.stdin)
        print(routine_id, file=backend.stdin)
        print(p_spectrum, file=backend.stdin)
        print(args.maxfev, file=backend.stdin)
        backend.stdin.flush()

        # Main loop, controlled by the lines printed by the backend.
        first_expno = True
        while True:
            # Read in what the backend has to say.
            line = backend.stdout.readline()

            # CASE 1 -- Optimisation has converged (or terminated with
            #           CostFunctionError)
            if line.startswith("optima:"):
                optima = line.split()[1:]
                # Get the optimiser message as well.
                opt_message = backend.stdout.readline()
                break

            # CASE 2 -- Values to put in an experiment
            elif line.startswith("values:"):
                # Increment expno if it's not the first time.
                if args.separate:
                    if not first_expno:
                        # Before doing anything, check first whether a dataset
                        # already exists, so that we don't overwrite
                        # anything...!
                        if next_expno_exists():
                            raise RuntimeError("Existing dataset found at "
                                               "next expno! poise has been "
                                               "terminated.")
                        else:
                            XCMD("iexpno")
                            RE(current_dataset)
                            RE_IEXPNO()
                            current_dataset = CURDATA()
                            XCMD("browse_update_tree")
                    else:
                        first_expno = False
                # Make sure we're at the correct dataset.
                RE(current_dataset)
                # Obtain the values and set them
                values = line.split()[1:]
                if len(values) != len(routine.pars):
                    raise RuntimeError("Invalid values passed from backend: "
                                       "'{}'".format(line))
                for value, par in zip(values, routine.pars):
                    try:
                        float(value)
                    except ValueError:
                        raise RuntimeError("Invalid values passed from "
                                           "backend: '{}'".format(line))
                    else:
                        convert_name_and_putpar(par, value)
                # Generate WaveMaker shapes if necessary
                if pulprog_contains_wvm():
                    XCMD("wvm -q")
                # Make sure we're at the correct dataset (again).
                RE(current_dataset)
                # Run acquisition and processing
                XCMD("xau {}".format(routine.au))
                # Check whether acquisition is complete
                if not acqu_done():
                    raise RuntimeError("Acquisition stopped prematurely. "
                                       "poise has been terminated.")
                # Tell backend script it's done
                print("done", file=backend.stdin)
                print(make_p_spectrum(), file=backend.stdin)
                backend.stdin.flush()

            # CASE 3 - Value of cost function at current spectrum
            # We take this value and put it inside the parameter TI.
            elif line.startswith("cf:"):
                cf_val = line.split()[1]
                PUTPAR("TI", cf_val)

            # CASE 4 - Traceback for backend error
            # The entire main() routine in the backend is wrapped by a
            # try/except which catches all exceptions and propagates them to
            # the frontend by printing the traceback.
            elif line.startswith("Backend exception: "):
                raise RuntimeError(line)

            # CASE 5 - Everything else
            else:
                raise RuntimeError("Invalid message from backend: '{}'."
                                   "Please see error log for more "
                                   "information.".format(line))
    # Cleanup code if anything goes wrong.
    except (Error, RuntimeError) as e:
        # Reset TI to empty string
        PUTPAR("TI", " ")
        # For some really silly reason, I can't just call
        # kill_remaining_backends() here, or else if the *original* POISE is
        # killed (using TopSpin's `kill`), it throws a
        # java.nio.BufferOverflowException. On Windows 10, this *does*
        # terminate the backend process, but it doesn't delete the .pidXXX
        # file. In principle there's nothing *wrong* with that because the
        # .pidXXX file will be cleaned up on the next run, but this just works
        # more cleanly.
        XCMD("xpy poise --kill")
        # BTW, err_exit() only gets called if it's RuntimeError. I don't know
        # why. If it's killed via TopSpin, then it just shows
        # java.lang.ThreadDeath as usual.
        err_exit("Error during acquisition loop:\n{}".format(e.message),
                 log=True)

    # Optimisation successfully ended, or prematurely ended?
    opt_success = opt_message == "Optimisation terminated successfully."

    # Store the optima in the (final) dataset, and show a message to the user
    # if not in quiet mode.
    RE(current_dataset)
    s = opt_message + "\n"
    for par, optimum in zip(routine.pars, optima):
        so_far = "" if opt_success else " (so far)"
        s = s + "Optimal value for {}{}: {}\n".format(par, so_far, optimum)
        if not args.separate:
            convert_name_and_putpar(par, optimum)
    if not args.separate:
        s = s + "These values have been set in the current experiment.\n"
    s = s + "Detailed information can be found at: {}".format(p_optlog)
    if not args.quiet:
        MSG_nonmodal(s, "POISE optimisation complete")
    EXIT()


class cst:
    """
    Constants used throughout the programme.
    """
    # TopSpin's internal codes for keypresses
    ENTER = -10
    ESCAPE = -27
    # List-type parameters. These parameters are accessed in the TopSpin
    # command line using "cnst1" (for example), but TopSpin's GETPAR() and
    # PUTPAR() Python functions need a space between the name and number, i.e.
    # "CNST 1".
    params_with_space = ["CNST", "D", "P", "PLW", "PLDB", "PCPD",
                         "GPX", "GPY", "GPZ", "SPW", "SPDB", "SPOFFS",
                         "SPOAL", "L", "IN", "INP", "PHCOR"]


def MSG_nonmodal(msg, title=""):
    """
    Non-modal version of MSG().

    For some reason TopSpin forces MSG() to be modal and if you want a
    non-modal message you have to use ERRMSG(), which has a scary-looking error
    sign, not exactly what I want people to see when they finish an
    optimisation!

    Well, it turns out you can reverse engineer their source code quite easily
    using inspect.getsource(), so this was born...

    Parameters
    ----------
    msg : str
        Message to be displayed.
    title : str, optional
        Title of the dialog box.
    """
    dialog = mfw.BInfo()
    dialog.setMessage(msg)
    dialog.setTitle(title)
    dialog.setBlocking(0)  # in ordinary MSG() this is 1
    dialog.show()


def err_exit(error, log=False):
    """
    Shows an error message in TopSpin, optionally logs it to the frontend error
    log, then quits the Python script.

    Parameters
    ----------
    error : str
        Error message to be displayed.
    log : bool, optional
        Whether to log the message to the frontend error log.

    Returns
    -------
    None
    """
    ERRMSG(title="poise", message=error)
    if log:
        current_dataset = CURDATA()
        p_front_err = os.path.normpath(os.path.join(current_dataset[3],
                                                    current_dataset[0],
                                                    current_dataset[1],
                                                    "poise_err_frontend.log"))
        with open(p_front_err, "a") as ferr:
            ferr.write("======= From frontend =======\n")
            ferr.write(datetime.now().strftime("%Y-%m-%d %H:%M:%S") + "\n")
            ferr.write(error + "\n\n")
    EXIT()


def list_files(path, ext="", with_ext=False):
    """
    Lists all files with a given extension found in a directory, if the
    directory exists.

    Parameters
    ---------
    path : str
        Directory to be searched in.
    ext : str, optional
        Extension of files to be searched for. Should include the dot, i.e. to
        search for Python files, use ext=".py".
    with_ext : bool, optional
        Whether to return filenames with the extension, or without it (i.e.
        with the extension stripped).

    Returns
    -------
    fnames : list of str
        A list of the filenames as strings.
    """
    if os.path.isdir(path):
        if ext == "":
            return [i for i in os.listdir(path)
                    if os.path.isfile(os.path.join(path, i))]
        else:
            if with_ext:
                return [i for i in os.listdir(path)
                        if i.endswith(ext)
                        and os.path.isfile(os.path.join(path, i))]
            else:
                return [os.path.splitext(i)[0] for i in os.listdir(path)
                        if i.endswith(ext)
                        and os.path.isfile(os.path.join(path, i))]
    else:
        return []


def run_now():
    """
    Asks the user whether they want to run an optimisation now, or just set up
    a new routine.

    Parameters
    ----------
    None

    Returns
    -------
    bool : True if user wants to run an optimisation now.
    """
    x = SELECT(title="poise",
               message=("Do you want to run the optimisation now?"),
               buttons=["Yes", "No"])
    return x in [0, cst.ENTER]


def get_routine_id():
    """
    Interactively prompts the user to choose a routine, using dialog boxes in
    TopSpin.

    Parameters
    ---------
    None

    Returns
    -------
    routine_id : str or None
        If a saved routine is chosen, the routine name is returned as a string
        (the corresponding routine .json file should be found as
        <routine_name>.json in .../py/user/poise_backend/routines).

        Returns None if the user requests a new routine.
    """
    # Check for existing saved routines
    saved_routines = list_files(p_routines, ext=".json")

    # Return immediately if there are no saved routines
    if saved_routines == []:
        return None
    # Otherwise, show the user a dialog box.
    else:
        x = SELECT(title="poise",
                   message=("Do you want to use a saved routine, or create a "
                            "new routine?"),
                   buttons=["Saved routine", "New routine"])

        # User pressed 'Saved routine', or Enter
        if x in [0, cst.ENTER]:
            s = "\n".join(saved_routines)
            # Prompt the user to choose one
            y = INPUT_DIALOG(title="poise: available routines",
                             header="Available routines:\n" + s,
                             items=["Routine:"])
            if y is None:  # User closed the dialog
                EXIT()
            elif y[0] in saved_routines:  # User gave a correct routine name
                return y[0]
            else:  # User wrote something, but it wasn't correct
                if y[0] != "":
                    err_exit("The routine '{}' was not found.".format(y[0]))
        # User pressed 'New routine'.
        elif x == 1:
            return None
        # User did something else, like Escape or close button.
        else:
            EXIT()


def get_new_routine():
    """
    Prompts the user for all the details required to make a new routine, using
    TopSpin dialog boxes.

    Doesn't serialise the routine!

    Parameters
    ---------
    None

    Returns
    -------
    Routine
        The Routine object created from the given parameters.
    """
    # Prompt for routine name.
    name = INPUT_DIALOG(title="poise: choose a name...",
                        header="Please choose a name to save this routine as:",
                        items=["Name:"])
    if name is None or name[0].strip() == "":
        EXIT()
    name = name[0].replace(".py", "").strip()

    # Prompt for parameters to be optimised.
    opt_pars = INPUT_DIALOG(title="poise: choose parameters...",
                            header=("Please select experimental parameters "
                                    "to be optimised (separated by commas or "
                                    "spaces):"),
                            items=["Parameters:"])
    # Quit if nothing is given
    if opt_pars is None or opt_pars[0].strip() == "":
        EXIT()
    opt_pars = opt_pars[0].replace(",", " ").split()
    opt_pars = [i for i in opt_pars if i]  # remove empty strings

    # Prompt for minimum, maximum, initial values, and tolerances..
    lbs = []
    ubs = []
    inits = []
    tols = []
    for i in opt_pars:
        settings = INPUT_DIALOG(title="poise: settings for {}".format(i),
                                header=("Please choose the minimum value, "
                                        "maximum value, initial value, and "
                                        "tolerance for {}:".format(i)),
                                items=["Minimum:", "Maximum:",
                                       "Initial value:", "Tolerance:"])
        if settings is None:
            EXIT()
        else:
            lbs.append(process_values(i, settings[0]))
            ubs.append(process_values(i, settings[1]))
            inits.append(process_values(i, settings[2]))
            tols.append(process_values(i, settings[3]))

    # Check if the values are sensible
    s = ""
    for par, lb, ub, init, tol in zip(opt_pars, lbs, ubs, inits, tols):
        if lb > ub:
            s = s + ("Please set the maximum value of {} to be larger "
                     "than the minimum value. "
                     "[Current values: min={}, max={}]\n").format(par, lb, ub)
        else:
            if init < lb or init > ub:
                s = s + ("Please set the initial value of {} to a value "
                         "to be between {} and {}. "
                         "[Current value: {}]\n").format(par, lb, ub, init)
            if tol > (ub - lb) or tol <= 0:
                s = s + ("Please set the tolerance of {} to a positive value "
                         "smaller than {}. "
                         "[Current value: {}]\n").format(par, (ub - lb), tol)
    if s:
        err_exit(s)

    # Prompt for cost function.
    # Search for existing cost functions
    saved_cfs = detect_costfunctions().keys()

    if saved_cfs == []:
        err_exit("No cost functions have been defined!\n"
                 "Please reinstall poise to get a default set, or define "
                 "your own cost function based on the documentation.")
    else:
        s = ", ".join(saved_cfs)
        x = INPUT_DIALOG(title=("poise: choose a cost function and"
                                " (optionally) an AU programme..."),
                         header="Available cost functions: " + s,
                         items=["Cost function:", "AU programme (optional):"])
        if x is None:  # user closed the dialog
            EXIT()
        elif x[0] in saved_cfs:
            cf, au = x[0], x[1]
        else:
            err_exit("Cost function {} was not found. Exiting...".format(x[0]))

    return Routine(name, opt_pars, lbs, ubs, inits, tols, cf, au)


def load_routine(routine_name):
    """
    Reads in a Routine serialised as JSON inside the p_routines directory.

    Parameters
    ----------
    routine_name : str
        The name of the routine. The routine is read from the
        p_routines/<routine_name>.json file.

    Returns
    -------
    Routine
        The Routine object.
    """
    with open(os.path.join(p_routines, routine_name + ".json"), "rb") as f:
        routine = Routine(**json.load(f))
    return routine


def check_routine(routine):
    """
    Checks that a Routine object is valid. Exits the programme if it isn't.

    Parameters
    ----------
    routine : Routine
        The Routine object to be checked.

    Returns
    -------
    None
    """
    try:
        (routine.name, routine.pars, routine.lb, routine.ub,
         routine.init, routine.tol, routine.cf, routine.au)
    except AttributeError:
        err_exit("The routine file is invalid.\n"
                 "Please delete it and recreate it from within poise.")


def routine_to_str(routine):
    """
    Generates a paragraph of text that summarises the parameters of a Routine.

    Parameters
    ----------
    routine : Routine
        The Routine to be described.

    Returns
    -------
    str
        The various parameters of the Routine in human-readable form.
    """
    d = routine._asdict()
    s = routine.name

    for key in sorted(list(d.keys())):
        s += "\n\t{:5}: {}".format(key, str(d[key]))
    return s


def list_routines_cfs():
    """
    Shows the user a list of cost functions and their docstrings, then routines
    and their parameters. Invoked via ``poise -l``.

    Parameters
    ----------
    None

    Returns
    -------
    None
    """
    # Get cost functions
    cfs = detect_costfunctions()
    text = "Available cost functions (and their descriptions)\n"
    text += "-------------------------------------------------\n"
    cf_strings = [" * {}\n    {}".format(function_name, docstring)
                  for function_name, docstring in cfs.items()]
    text += "\n\n".join(cf_strings)
    # Get routines
    text += "\n"
    text += "\nAvailable routines\n-------------------\n"
    routine_names = sorted(list_files(p_routines, ext=".json"))
    routine_strings = [routine_to_str(load_routine(routine_name))
                       for routine_name in routine_names]
    text += "\n\n".join(routine_strings)
    VIEWTEXT(title="Available poise cost functions and routines", text=text)
    EXIT()


def detect_costfunctions():
    """
    Gets poise_backend/get_cfs.py to parse the costfunctions.py file (using the
    ast module) and pass a dictionary of defined cost functions and their
    docstrings back to the frontend.

    Parameters
    ----------
    None

    Returns
    -------
    available_cfs : dict
        Dictionary where keys are the names of the cost function and values are
        their docstrings.
    """
    p_get_cfs = os.path.join(p_poise, "get_cfs.py")
    p = subprocess.Popen([p_python3, p_get_cfs], stdout=subprocess.PIPE)
    fdict_as_str, _ = p.communicate()
    return json.loads(fdict_as_str)


def check_python3path():
    """
    Checks that the global variable p_python3 is set to a valid file. Exits the
    programme if it isn't.

    Parameters
    ----------
    None

    Returns
    -------
    None
    """
    if not os.path.isfile(p_python3):
        err_exit("The python3 executable was not found.\n"
                 "Please specify p_python3 in poise.py.")


def convert_name(name):
    """
    Converts the name of a parameter from the command-line accessible version
    (e.g. "cnst1") to the version that is suitable for the TopSpin GETPAR() and
    PUTPAR() functions (e.g. "CNST 1").

    Parameters
    ----------
    name : str
        Name of the parameter.

    Returns
    -------
    ts_name : str
        Name of the parameter that can be used with GETPAR() and PUTPAR().
    """
    # GETPAR() accepts: CNST 1, P 1, D 1, PLW 1, PLdB 1 (note small d), GPZ 1
    #                   SPW 1, SPdB 1, SPOFFS 1, SPOAL 1,
    #               but O1, O1P, SFO1 (note no space)
    ts_name = name.upper()
    ts_namel = ts_name.rstrip("1234567890")   # the word
    ts_namer = ts_name[len(ts_namel):]        # the number
    # Add a space if needed
    if ts_namel in cst.params_with_space:
        ts_name = ts_namel + " " + ts_namer
    # Convert DB back to dB
    ts_name = ts_name.replace("PLDB", "PLdB").replace("SPDB", "SPdB")
    return ts_name


def convert_name_and_getpar(name):
    """
    Obtains the value of an acquisition parameter. The parameter name is
    first converted into a format suitable for TopSpin's GETPAR() function.
    Returns the value as a float. Exits with an error if the value can't be
    converted to a float.

    Parameters
    ----------
    name : str
        Name of the parameter to be looked up, e.g. "cnst1" or "d10".

    Returns
    -------
    val : float
        Value of the input parameter.
    """
    # Get the value.
    val = GETPAR(convert_name(name))
    # Check that it's a float.
    try:
        val = float(val)
        return val
    except ValueError:
        err_exit("The acquisition parameter {} could not be found, "
                 "or does not have a numerical value.".format(name))


def convert_name_and_putpar(name, val):
    """
    Obtains the value of an acquisition parameter. The parameter name is
    first converted into a format suitable for TopSpin's GETPAR() function.
    Exits with an error if the value isn't a float.

    Parameters
    ----------
    name : str
        Name of the parameter to be looked up, e.g. "cnst1" or "d10".
    val : float
        The value to be set.

    Returns
    -------
    None
    """
    try:
        val = float(val)
        PUTPAR(convert_name(name), str(val))
    except ValueError:
        err_exit("The value {} for parameter {} "
                 "was invalid.".format(val, name))


def process_values(parname, input_string):
    """
    Does some processing on user input to allow familiar hacks, like
    entering 2m for a duration or 2k for 2048.

    Parameters
    ----------
    parname : str
        String containing the parameter name (either "p1" or "P 1" works).
    input_string : str
        The string entered by the user into the dialog box.

    Returns
    -------
    float
        The processed value, as a float.
    """
    # We need to determine whether the parameter is a pulse (for which the
    # default units are us), or a delay (for which the default units are
    # seconds).
    pulse = 1e6 if parname.upper().rstrip("1234567890").strip() == "P" else 1
    try:
        if input_string.endswith("k"):
            f = int(input_string[:-1])  # int only, can't have "2.5k"
            return 1024*f
        elif input_string.endswith("s"):
            f = float(input_string[:-1])
            return f*pulse
        elif input_string.endswith("m"):
            f = float(input_string[:-1])
            return f*1e-3*pulse
        elif input_string.endswith("u"):
            f = float(input_string[:-1])
            return f*1e-6*pulse
        else:
            f = float(input_string)
            return f
    except ValueError:
        err_exit("'{}' was not a valid input for {}. "
                 "Please try again.".format(input_string, parname))


def make_p_spectrum():
    """
    Constructs the path to the procno folder of the active spectrum.

    Parameters
    ----------
    None

    Returns
    -------
    str
        The absolute path to the procno folder.
    """
    x = CURDATA()
    p = os.path.join(x[3], x[0], x[1], "pdata", x[2])
    return p


def next_expno_exists():
    """
    Checks whether the expno folder with expno 1 greater than the current
    experiment exists.

    Parameters
    ---------
    None

    Returns
    -------
    bool
        True if it does. False if not.
    """
    x = CURDATA()
    next_expno = str(int(x[1]) + 1)
    p = os.path.join(x[3], x[0], next_expno)
    return os.path.isdir(p)


def acqu_done():
    """
    Checks whether the spectrum was completely acquired. Useful for stopping
    poise if acquisition is prematurely terminated (e.g. by ``stop``).

    Parameters
    ----------
    None

    Returns
    -------
    bool
        True if the acquisition was successfully completed. False otherwise.
    """
    # Check NS
    if GETPAR("NS") != GETPARSTAT("NS"):
        return False
    # Check TD (see #25)
    if GETPAR("TD") != GETPARSTAT("TD"):
        return False
    # Check indirect dimension TD for 2Ds
    if GETACQUDIM() > 1:
        if GETPAR("TD", axis=1) != GETPARSTAT("TD", axis=1):
            return False
    # OK, looks like the acquisition did complete
    return True


def au_exists(au):
    """
    Checks for the existence of a named AU programme.

    Parameters
    ----------
    au : str
        Name of the AU programme.

    Returns
    -------
    bool : Whether the AU programme exists.
    """
    # Undocumented TopSpin function. Fun.
    au_folders = getParfileDirs(2)
    # Just check whether it exists in each of the folders.
    for folder in au_folders:
        if os.path.exists(os.path.join(folder, au)):
            return True
    return False


def pulprog_contains_wvm():
    """
    Parses the text of the current pulse programme to see whether WaveMaker
    directives are used.

    Parameters
    ----------
    None

    Returns
    -------
    bool
        True if the pulse programme does contain WaveMaker directives. False
        otherwise.
    """
    ppname = GETPAR("PULPROG")

    # Get path to the pulse programme text
    ppdirs = getParfileDirs(0)
    for ppdir in ppdirs:
        if ppname in os.listdir(ppdir):
            fname = os.path.join(ppdir, ppname)
            break

    # Search for lines containing WaveMaker directives
    wvm_rgx = re.compile(r"^\s*;[A-Za-z0-9]+:wvm:")
    with open(fname, "r") as file:
        for line in file:
            if re.search(wvm_rgx, line):
                return True
    return False


def kill_remaining_backends():
    """
    Checks the poise_backend folder for any .pidXXXX files (which represent
    backends that have not exited), then kills all the associated PIDs and
    deletes the files.
    """
    for file in os.listdir(p_poise):
        if file.startswith(".pid"):
            try:
                pid = int(file[4:])
            except ValueError:  # not an int
                pass
            else:
                kill_pid(file[4:])
                os.remove(os.path.join(p_poise, file))


def kill_pid(pid):
    """
    Kills a process with PID pid.

    Parameters
    ----------
    pid : int or str
        The process ID.

    Returns
    -------
    None
    """
    from java.lang.System import getProperty
    # Windows
    if "wind" in getProperty("os.name").lower():
        subprocess.call(["taskkill", "/f", "/pid", str(pid)])
    # *nix
    else:
        subprocess.call(["kill", "-9", str(pid)])


def create_routine(cli_args):
    """
    Creates a .json file for a one-parameter routine.

    Parameters
    ----------
    cli_args : list of str
        This list should contain 8 strings, namely: name, parameter, minimum
        value, maximum value, initial value, tolerance, cost function, and AU
        programme.

    Returns
    -------
    None
    """
    name, param, minval, maxval, init, tol, cf, au = cli_args
    try:
        minval = float(minval)
        maxval = float(maxval)
        init = float(init)
        tol = float(tol)
    except ValueError:
        err_exit("poise --create: invalid input")
    if (minval > init) or (init > maxval) or (tol > (maxval - minval)):
        err_exit("poise --create: invalid input")
    # If we got here, the input is probably OK.
    routine = Routine(name, [param], [minval], [maxval], [init], [tol], cf, au)
    with open(os.path.join(p_routines, routine.name + ".json"), "wb") as f:
        json.dump(routine._asdict(), f)
    EXIT()


def delete_routine(name):
    """
    Deletes a routine .json file.

    Parameters
    ----------
    name : str
        Name of the routine to be deleted.

    Returns
    -------
    None
    """
    # Check whether said routine really exists
    saved_routines = list_files(p_routines, ext=".json")
    if name not in saved_routines:
        err_exit("The routine {} was not found.".format(name))
    else:
        os.remove(os.path.join(p_routines, name + ".json"))
        MSG_nonmodal("Routine {} deleted.".format(name))
        EXIT()


def subinstall(target):
    """
    Installs the necessary AU scripts / Python scripts / routines for p1 and
    DOSY optimisations. This simply invokes a Python 3 script which does all
    the real work.

    Parameters
    ----------
    target : str from {"p1", "dosy"}
        The set of scripts/routines to install.
         - For "p1", installs the poisecal AU programme only. The routine is
           set up by the user according to the instructions in the
           documentation.
         - For "dosy", installs the dosy_opt Python programme as well as the
           dosy and dosy_aux routines required to run that optimisation.
        In both cases, the cost functions are already pre-installed with POISE.

    Returns
    -------
    None
    """
    if target not in ["p1", "dosy"]:
        err_exit("poise --install only takes the arguments 'p1' or 'dosy'.")
    else:
        # Delegate to external script.
        p_poise_install = os.path.join(p_poise,
                                       "_poise_install",
                                       "_poise_install.py")
        subprocess.Popen([p_python3, p_poise_install, target])
        MSG_nonmodal("poise --install {}: done\n".format(target))
        EXIT()


class TSArgumentParser(argparse.ArgumentParser):
    """
    Custom ArgumentParser class which shows a TopSpin dialog box with errors
    instead of printing to stderr.
    """
    def print_usage(self, file=None):
        VIEWTEXT(title="poise usage", text=self.format_usage())
        EXIT()

    def print_help(self, file=None):
        VIEWTEXT(title="poise help", text=self.format_help())
        EXIT()

    def error(self, message):
        ERRMSG(title="poise", message=message)
        EXIT()


if __name__ == "__main__":
    # Parse arguments.
    parser = TSArgumentParser(
        prog="poise",
        description=("Python script for optimisation of acquisition "
                     "parameters in TopSpin. Requires a separate Python 3 "
                     "backend (run `pip install nmrpoise` to download the "
                     "backend). Full documentation can be found at "
                     "https://foroozandehgroup.github.io/nmrpoise.")
    )
    # Commands that don't actually do any optimisations; they just run one
    # function and exit. These are mutually exclusive (since in general the
    # user should only be doing one of these at a time).
    me_group = parser.add_mutually_exclusive_group()
    parser.add_argument(
        "routine",
        nargs="?",
        help=("The name of the routine to use (use 'poise --list' to see "
              "available routines).")
    )
    parser.add_argument(
        "-a",
        "--algorithm",
        default="nm",
        choices=["nm", "mds", "bobyqa"],
        help="Optimisation algorithm to use. (default: 'nm')"
    )
    me_group.add_argument(
        "--create",
        type=str,
        nargs=8,
        help=("Create a new POISE routine for optimisation of one parameter."
              " For multiple parameter optimisations, please use the GUI."
              " Invoke using 'poise --create [NAME] [PARAM] [MINVAL] [MAXVAL]"
              " [INITVAL] [TOLS] [CF] [AU]. Note that unlike in the GUI, the"
              " AU programme must be specified here. Also, shorthand notation"
              " such as 'u' or 'm' for units are not allowed here.")
    )
    me_group.add_argument(
        "--delete",
        type=str,
        help=("Delete an existing POISE routine.")
    )
    me_group.add_argument(
        "--install",
        choices=["p1", "dosy"],
        help=("Ultraquick installation. Refer to documentation for more info.")
    )
    me_group.add_argument(
        "--kill",
        action="store_true",
        help=("Kill POISE backends that may still be running.")
    )
    me_group.add_argument(
        "-l",
        "--list",
        action="store_true",
        help="List all available routines and exit."
    )
    parser.add_argument(
        "--maxfev",
        type=int,
        default=0,
        help=("Maximum function evaluations to allow. Use 0 to not enforce "
              "any limit (technically there is a hard limit, which is 500 "
              "times the number of parameters being optimised). (default: 0)")
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help=("Don't display any messages at the end of the optimisation. "
              "Using this flag is necessary if POISE is to be run under "
              "automation. (default: off)")
    )
    parser.add_argument(
        "-s",
        "--separate",
        action="store_true",
        help=("Use separate expnos for each function evaluation. (default: "
              "off)")
    )
    args = parser.parse_args()

    # List
    if args.create:
        create_routine(args.create)
    if args.delete:
        delete_routine(args.delete)
    elif args.install:
        subinstall(args.install)
    elif args.list:
        list_routines_cfs()
    elif args.kill:
        kill_remaining_backends()
    else:
        main(args)



"""
shared.py
---------

Stores the _g class which holds global variables (to track the state of the
optimisation).

SPDX-License-Identifier: GPL-3.0-or-later
"""

from pathlib import Path


class _g():
    """
    Class to store the "global" variables.

    Attributes
    ----------
    optimiser : str from {'nm', 'mds', 'bobyqa'}
        The optimiser being used.

    routine_id : str
        The name of the routine being used.

    p_spectrum : |Path|
        The path to the procno folder of the spectrum just acquired. (e.g.
        ``/path/to/data/1/pdata/1``)

    p_optlog : |Path|
        The path to the currently active ``poise.log`` file.

    p_errlog : |Path|
        The path to the currently active ``poise_err_backend.log`` file.

    maxfev : int
        The maximum number of function evaluations specified by the user. Can
        be zero, indicating no limit (beyond the hard limit of 500 times the
        number of parameters).

    p_poise : |Path|
        The path to the ``$TS/exp/stan/nmr/py/user/poise_backend`` folder.

    spec_f1p : float or tuple of float
        The ``F1P`` parameter. For a 1D spectrum this is a float. For a 2D
        spectrum this is a tuple of floats (``indirect, direct``) corresponding
        to the values of ``F1P`` in both spectral dimensions.

    spec_f2p : float or tuple of float
        The ``F2P`` parameter.
    """
    optimiser = None
    routine_id = None
    p_spectrum = None
    p_optlog = None
    p_errlog = None
    maxfev = 0
    nfev = 0
    p_poise = Path(__file__).parent.resolve()
    spec_f1p = None
    spec_f2p = None



"""
costfunctions_user.py
---------------------

For user-defined cost functions.

Any cost functions defined here will shadow cost functions defined in
costfunctions.py. That means that (for example) if you redefine minabsint()
here, it will take precedent over the minabsint() defined in costfunctions.py.

SPDX-License-Identifier: GPL-3.0-or-later
"""

# You can import any other packages you need here (as long as they are present
# on the Python 3 used to install POISE).
import numpy as np

# Keep these imports here -- they provide the helper functions described in the
# documentation.
from .cfhelpers import *
from .shared import _g


# You're free to do anything here! For (very thoroughly commented) examples of
# how to setup your own cost functions, please check out costfunctions.py.



"""
_poise_install.py
-----------------

Tiny script which handles the `poise --install` commands from the frontend.

SPDX-License-Identifier: GPL-3.0-or-later
"""

import sys
from shutil import copy2
from pathlib import Path


def main():
    this_dir = Path(__file__).parent.resolve().expanduser()
    # poise --install p1: copy poisecal to the AU directory. The cost function
    # is installed by default.
    if sys.argv[1] == "p1":
        src = this_dir / "poisecal"
        dst = this_dir.parents[3] / "au" / "src" / "user"
        copy2(src, dst)
    # poise --install dosy: copy dosy_opt.py to the PY directory. The cost
    # function is already installed, as well.
    elif sys.argv[1] == "dosy":
        # dosy_opt.py
        src = this_dir / "dosy_opt.py"
        dst = this_dir.parents[1]
        copy2(src, dst)
    else:
        print("sys.argv[1] not provided, exiting")


if __name__ == "__main__":
    main()



"""
make_double_saltire.py
----------------------

Download from: https://git.io/JUHOZ

This script reads in a pulse duration (stored in us as P40) and bandwidth
(stored in Hz as CNST21), then creates a double saltire shaped file called
"auto_double_saltire" which can be directly used as the shaped file in
PSYCHE experiments.

This can be included in an AU programme using the line:

    XCMD("sendgui xpy make_double_saltire")

SPDX-License-Identifier: GPL-3.0-or-later
"""

from __future__ import division
from math import sin, cos, pi, sqrt, degrees, atan2


def make_double_saltire(duration, bandwidth):
    """
    Parameters
    ----------
    duration : float
        Time in units of seconds.

    bandwidth : float
        Bandwidth in units of Hz.

    Returns
    -------
    A : list of float
        Amplitudes, scaled between 0 and 100.
    phi : list of float
        Phases, in degrees between -180 and 180.

    These lists can be directly passed to the Bruker SAVE_SHAPE routine, as
    demonstrated in this script. The amplitude of the pulse should be
    automatically calculated in the pulse programme.
    """
    # Our strategy is to generate the first saltire first.

    # Number of points in half of the pulse. We use 1 point per us.
    npoints = int(duration * 1000000 / 2)
    # 20% smoothing.
    s = 0.2
    # This is equivalent to ls = np.linspace(0, 1, npoints), but you can't use
    # numpy in TopSpin...
    ts = [i/(npoints - 1) for i in range(0, npoints)]

    # Construct the amplitude piecewise function for a single chirp.
    A_chirp = [sin((pi * t) / (2 * s)) if t < s
               else 1 if s <= t and t <= 1 - s
               else sin((pi * (1 - t)) / (2 * s))
               for t in ts]
    # These are phases for a single chirp, not a saltire.
    phi_chirp = [pi * (duration/2) * bandwidth * ((t - 0.5) ** 2) for t in ts]
    # Calculate x-coefficients (Cx) for the saltire.
    Cx = [100 * a * cos(phi) for a, phi in zip(A_chirp, phi_chirp)]
    # The Cy's are 100 * a * sin(phi) ..., but that's only for one chirp. For a
    # saltire, the two component chirps have equal and opposite Cy's (because
    # sin(phi) = -sin(-phi)), so the total Cy is simply zero for every point.
    # In principle we could make a list full of zeroes, like Cy = [0 for _ in
    # range(npoints)], but that's a waste of memory and time.
    # Incidentally, the Cx's for both chirps add up since cos(phi) = cos(-phi),
    # but because the amplitudes are scaled to [0, 100], there's no point
    # actually adding them up.

    # Convert back to polar coordinates. Ordinarily A = sqrt(xi * xi + yi *
    # yi), but since all the yi's are zero, we just leave it out of the
    # equation.
    A = [sqrt(xi * xi) for xi in Cx]
    # Likewise here we use atan2(0, xi) instead of atan2(yi, xi).
    phi = [degrees(atan2(0, xi)) for xi in Cx]
    # Finally, we double the lists (because it's a double saltire).
    return A * 2, phi * 2


if __name__ == "__main__":
    # Let the user know what's going on
    SHOW_STATUS("make_double_saltire: generating saltire...")
    # Pulses are by default in us, so must convert first
    duration = float(GETPAR("P 40"))/1000000
    # Bandwidth is specified in Hz.
    bandwidth = float(GETPAR("CNST 21"))
    # Generate the amplitudes and phases.
    A, phi = make_double_saltire(duration, bandwidth)
    # Save it to disk.
    shape_name = "auto_double_saltire"
    SAVE_SHAPE(shape_name, "Excitation", A, phi)
    # Use it as SPNAM40. It's too easy to forget this!
    PUTPAR("SPNAM 40", shape_name)



"""
SPDX-License-Identifier: GPL-3.0-or-later
"""

import sys
import subprocess
from pathlib import Path
from setuptools import setup, find_packages
from setuptools.command.install import install


class TopSpinInstall(install):
    def run(self):
        p = Path(__file__).parent.resolve()
        p_install = p / "nmrpoise" / "topspin_install.py"
        subprocess.run([sys.executable, str(p_install)], check=True)
        super().run()


class noTopSpinInstall(install):
    """Doesn't install to TopSpin. Useful for testing."""
    pass


with open("README.md", "r") as fp:
    long_description = fp.read()

# Read the version number in as __version__
exec(open('nmrpoise/_version.py').read())

setup(
    name="nmrpoise",
    version=__version__,
    author="Jonathan Yong",
    author_email="jonathan.yong@chem.ox.ac.uk",
    description=("Parameter Optimisation by Iterative Spectral Evaluation, "
                 "a TopSpin-compatible NMR package"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/foroozandehgroup/nmrpoise",
    packages=[
        "nmrpoise",
        "nmrpoise.poise_backend",
        "nmrpoise.py",
        "nmrpoise.au",
    ],
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=["numpy>=1.17.0",
                      "Py-BOBYQA",
                      "pandas"
                      ],
    cmdclass={"install": TopSpinInstall,
              "notopspin": noTopSpinInstall,
    },
    entry_points={
        "console_scripts": [
            "poise_ts = nmrpoise.topspin_install:main",
            "poise_addons = nmrpoise.topspin_install:install_addons",
        ]
    },
)



"""
topspin_install.py
------------------

Script which carries out installation of POISE files to TopSpin directories.

SPDX-License-Identifier: GPL-3.0-or-later
"""

import os
import sys
from shutil import copy2 as cp
import platform
from pathlib import Path
from glob import glob


def cp_r(src, dest):
    """
    Recursively copies src to dest.

    This is a reimplementation of shutil.copytree(dirs_exist_ok=True), because
    the dirs_exist_ok keyword argument is not available in Python <=3.7. The
    ignore pattern is hardcoded.

    Modified from https://stackoverflow.com/a/15824216/7115316.
    """
    if os.path.isdir(src):
        if not os.path.isdir(dest):
            os.makedirs(dest)
        files = os.listdir(src)
        for f in files:
            if f not in ["costfunctions_user.py"]:
                cp_r(os.path.join(src, f), os.path.join(dest, f))
    else:
        cp(src, dest)


def get_ostype():
    """
    Attempts to find out the operating system that it's running on. This
    influences the glob pattern used when searching for the TopSpin directory,
    as well as the error message that is shown.

    Parameters
    ----------
    None

    Returns
    -------
    ostype: str
        "unix" or "win".
    """
    osname = platform.system()
    if osname in ["Darwin", "Linux"]:
        return "unix"
    elif osname in ["Windows"]:
        return "win"
    else:
        raise OSError("Unsupported operating system. "
                      "Please perform a manual installation.")


def main():
    """
    Attempts to install poise's core scripts (frontend and beckend) to TopSpin
    directory. This function also makes sure to replace p_python3 (the path to
    the Python 3 executable) in the frontend script.

    Parameters
    ----------
    None

    Returns
    -------
    None
    """
    dirname = Path(__file__).parent.resolve()
    ostype = get_ostype()
    ts_paths = get_topspin_path(ostype)

    # Set path to Python executable in poise.py
    poise_py = dirname / "poise.py"
    poise_backend = dirname / "poise_backend"
    poise_py_out = dirname / "poise.py.out"
    # Iterate over lines in poise.py and print them to poise.py.out
    # Essentially this is sed -i behaviour.
    # open() only takes Path objects in >= 3.6
    with open(str(poise_py), "r") as infile, \
            open(str(poise_py_out), "w") as outfile:
        for line in infile:
            if line.startswith("p_python3 = "):
                line = f"p_python3 = r\"{Path(sys.executable)}\""
            # Use rstrip() to remove extra newlines since print() already
            # includes one for us.
            print(line.rstrip(), file=outfile)
    # Copy the new file over the old file.
    cp(str(poise_py_out), str(poise_py))

    # Copy the files to TopSpin's directories.
    for ts_path in ts_paths:
        ts_user_path = ts_path / "py" / "user"
        ts_backend_path = ts_path / "py" / "user" / "poise_backend"
        # Copy poise.py and poise_backend/ to TopSpin path, except for
        # costfunctions_user.py (which is hardcoded as an exception in cp_r()).
        cp(str(poise_py), str(ts_user_path))
        cp_r(str(poise_backend), str(ts_backend_path))
        # Copy costfunctions_user.py but only if it's not found.
        cf_user_path = poise_backend / "costfunctions_user.py"
        ts_cf_user_path = ts_backend_path / "costfunctions_user.py"
        if not ts_cf_user_path.exists():
            cp(str(cf_user_path), str(ts_cf_user_path))
        # Copy core AU scripts over. The others can be installed with
        # poise_addons.
        ts_au_user_path = ts_path / "au" / "src" / "user"
        core_au_scripts = ["poise_1d", "poise_2d"]
        for auscript in core_au_scripts:
            src = dirname / "au" / auscript
            dest = ts_au_user_path / auscript
            cp(str(src), str(dest))
    print("topspin_install.py: completed")


def install_addons():
    """
    Attempts to install additional AU programmes and Python scripts into the
    TopSpin user directories.

    This function can be called using the ``poise_addons`` command-line entry
    point.

    Parameters
    ----------
    None

    Returns
    -------
    None
    """
    dirname = Path(__file__).parent.resolve()
    ostype = get_ostype()
    ts_paths = get_topspin_path(ostype)

    # List of all the files we want to copy over.
    python_scripts = [
        "make_double_saltire.py",
        "dosy_opt.py",
        "poisecalpy.py",
    ]
    au_scripts = [
        "poise_psyche",
        "poise_2pdosy",
        "poisecal",
    ]
    pulse_programmes = [
        "asaphsqc_jy",
        "doneshot_1d_jy",
        "doneshot_2d_jy",
        "psyche_1dopt_jy",
        "spfgsenoezs2.1_jy",
    ]
    # Copy them to TopSpin's directories.
    for ts_path in ts_paths:
        # Python scripts first
        ts_py_user_path = ts_path / "py" / "user"
        for pyscript in python_scripts:
            src = dirname / "py" / pyscript
            dest = ts_py_user_path / pyscript
            cp(str(src), str(dest))
        # AU scripts next
        ts_au_user_path = ts_path / "au" / "src" / "user"
        for auscript in au_scripts:
            src = dirname / "au" / auscript
            dest = ts_au_user_path / auscript
            cp(str(src), str(dest))
        # Pulse programmes finally
        pp_user_path = ts_path / "lists" / "pp" / "user"
        for pulprog in pulse_programmes:
            src = dirname / "pp" / pulprog
            dest = pp_user_path / pulprog
            cp(str(src), str(dest))
    print("install_addons: complete")


def get_topspin_path(ostype):
    """
    Searches for the path to the TopSpin /exp/stan/nmr folder(s).

    First searches the environment variable TSDIR and returns that if it is a
    valid path. Otherwise, tries to find the path to the most recent TopSpin
    directory.  Searches under /opt on Unix/Linux systems, or under C:\\Bruker
    on Windows.

    Parameters
    ----------
    ostype : str
        "unix" or "win". The string returned by get_ostype().

    Returns
    -------
    tsdirs : list of pathlib.Path
        Every /exp/stan/nmr directory found. This can be more than one if there
        is more than one version of TopSpin installed.

    Raises
    ------
    RuntimeError
        If no paths were found. We need to throw an error so that the pip
        installation fails.
    """
    invalid_envvar_error = (
        "The TopSpin installation directory was specified as the environment "
        "variable TSDIR, but it was not a valid path.\n\n"
        "Please make sure that TSDIR points to the ../exp/stan/nmr folder in "
        "TopSpin."
    )
    no_tsdir_unix_error = (
        "A valid TopSpin installation directory was not found.\n"
        "Please set it using the TSDIR environment variable:\n\n"
        "\texport TSDIR=/opt/topspinX.Y.Z/exp/stan/nmr\n"
    )
    no_tsdir_win_error = (
        "A valid TopSpin installation directory was not found.\n"
        "Please set it using the TSDIR environment variable:\n\n"
        "\t$env:TSDIR = \"C:\\Bruker\\TopSpinX.Y.Z\\exp\\stan\\nmr\"\n"
    )
    if "TSDIR" in os.environ:
        ts = os.environ["TSDIR"]    # KeyError if not provided
        py_user = ts / "py" / "user"
        if not ts.is_dir() and py_user.is_dir():
            raise RuntimeError(invalid_envvar_error)
        dirs = [ts]
    else:
        if ostype == "unix":
            glob_query = "/opt/topspin*/exp/stan/nmr"
        elif ostype == "win":
            glob_query = r"C:\Bruker\TopSpin*\exp\stan\nmr"
        dirs = glob(glob_query)

        if len(dirs) == 0:              # No TopSpin folders found
            if ostype == "unix":
                raise RuntimeError(no_tsdir_unix_error)
            if ostype == "win":
                raise RuntimeError(no_tsdir_win_error)
        else:
            # Convert to pathlib.Path and filter out invalid entries
            dirs = [Path(dir) for dir in dirs
                    if (Path(dir) / "py" / "user").is_dir()]
            # Error out if no valid ones were found
            if dirs == []:
                if ostype == "unix":
                    raise RuntimeError(no_tsdir_unix_error)
                if ostype == "win":
                    raise RuntimeError(no_tsdir_win_error)

    print("topspin_install.py: found TopSpin paths", dirs)
    return dirs


if __name__ == "__main__":
    main()



"""
cfhelpers.py
------------

Script containing functions to read spectra, etc. which can be used in cost
functions.

SPDX-License-Identifier: GPL-3.0-or-later
"""

from pathlib import Path

import numpy as np

from .shared import _g


class CostFunctionError(Exception):
    """
    Custom exception class which can be raised *inside* a cost function in
    order to stop the optimisation immediately.

    For the Nelder-Mead and MDS optimisers, the optimisation will still return
    the best value found so far. For the BOBYQA optimiser, no value will be
    returned.

    If raised directly with ``CostFunctionError(str)``, the current iteration
    (i.e. the iteration on which it was raised) will *not* be included in the
    optimisation results. If you wish to include the current iteration, then
    pass the value of the cost function as the second parameter, viz.
    ``CostFunctionError(str, cf_val=1e6)``.

    Examples
    --------
    One might envision a scenario in which *any* value of the cost function
    below a certain threshold is considered 'good enough' to be the optimum.
    To obtain this effect, write a cost function with the following structure:

    >>>def cost_function():
    >>>    cf_val = foo()   # whatever calculation you want here
    >>>    if cf_val < threshold:
    >>>       raise CostFunctionError("Cost function was below threshold.",
    >>>                               cf_val=cf_val)
    >>>    return cf_val
    """
    def __init__(self,
                 message=("The cost function raised an error."),
                 cf_val=None):
        self.message = message
        self.cf_val = cf_val


def make_p_spec(path=None, expno=None, procno=None):
    """
    Constructs a |Path| object corresponding to the procno folder
    ``<path>/<expno>/pdata/<procno>``. If parameters are not passed, they are
    inherited from the currently active spectrum (_g.p_spectrum).

    Thus, for example, ``make_p_spec(expno=1, procno=1)`` returns a path to the
    spectrum with EXPNO 1 and PROCNO 1, but with the same name as the currently
    active spectrum.

    Parameters
    ----------
    path : str or |Path|, optional
        Path to the main folder of the spectrum (one level above the expno
        folders).
    expno : int, optional
    procno : int, optional

    Returns
    -------
    p_spec : |Path|
        Path pointing to the requested spectrum.
    """
    # this check is only needed for the unit tests, during normal usage it
    # should never be None
    if _g.p_spectrum is not None:
        default_procno = int(_g.p_spectrum.name)
        default_expno = int(_g.p_spectrum.parents[1].name)
        default_path = _g.p_spectrum.parents[2]
    # overwrite defaults with user-passed parameters
    if path is not None:
        path = Path(path)
    else:
        path = default_path
    expno = expno or default_expno
    procno = procno or default_procno
    return path / str(expno) / "pdata" / str(procno)


def _parse_bounds(b):
    """
    Parses the bounds string "lower..upper", or the bounds tuple (lower,
    upper), to get the lower and upper bounds.

    Parameters
    ----------
    b : str or tuple
        The bounds string "lower..upper", or the tuple (lower, upper).

    Returns
    -------
    lower : float or None
        Lower bound. None if not specified.
    upper : float or None
        Upper bound. None if not specified.
    """
    try:
        if isinstance(b, str):
            if b == "":
                return None, None
            elif b.startswith(".."):   # "..5" -> (None, 5)
                return None, float(b[2:])
            elif b.endswith(".."):   # "3.." -> (3, None)
                return float(b[:-2]), None
            elif ".." in b:
                x, y = b.split("..")
                return float(x), float(y)
            else:
                raise ValueError
        else:  # not a string, hopefully a tuple
            if len(b) != 2:
                raise ValueError
            # ValueError if b[0] isn't numeric
            lower = float(b[0]) if b[0] is not None else None
            upper = float(b[1]) if b[1] is not None else None
            return lower, upper
    except (ValueError, TypeError):
        raise ValueError(f"Invalid value {b} provided for bounds.") from None


def _ppm_to_point(shift, axis=None, p_spec=None):
    """
    Round a specific chemical shift to the nearest point in the spectrum.

    For unknown reasons, this does not correlate perfectly to the "index" that
    is displayed in TopSpin. However, the difference between the indices
    calculated here and the index in TopSpin is on the order of 1e-4 to 1e-3
    ppm. (More precisely, it's ca. 0.25 * SW / SI.)

    Parameters
    ----------
    shift : float
        Desired chemical shift.
    axis : int, optional
        Axis along which to calculate this. For 1D spectra this should be left
        as None. For 2D spectra, axis=0 and axis=1 correspond to the f1 and f2
        dimensions respectively.

    Returns
    -------
    int or None
        The desired point. None if the chemical shift lies outside the spectral
        window.
    """
    p_spec = p_spec or _g.p_spectrum
    si = getpar("SI", p_spec)
    o1p = getpar("O1", p_spec) / getpar("SFO1", p_spec)
    sw = getpar("SW", p_spec)

    # Pick out the appropriate value according to the relevant axis
    if axis is not None:
        if axis in [0, 1]:
            si, o1p, sw = si[axis], o1p[axis], sw[axis]
        else:
            raise ValueError(f"Invalid value '{axis}' for axis.")

    # Make sure it's within range
    highest_shift = o1p + 0.5*sw
    lowest_shift = o1p - 0.5*sw
    if shift > highest_shift or shift < lowest_shift:
        return None

    # Calculate the value
    spacing = (highest_shift - lowest_shift)/(si - 1)
    x = 1 + round((highest_shift - shift)/spacing)
    return int(x)


def get1d_fid(p_spec=None):
    """
    Returns the FID as a |ndarray|.

    Note that this does *not* modify the "group delay" at the beginning of the
    FID.

    Also, Bruker spectrometers record real and imaginary points in a sequential
    fashion. Therefore, each imaginary point in the ndarray is actually
    measured ``DW`` *after* the corresponding real point. When Fourier
    transforming, this can be accounted for by using fftshift().

    Parameters
    ----------
    p_spec : |Path|, optional
        Path to the procno folder of interest. (The FID is taken from the expno
        folder two levels up.) Leave blank to use the currently active spectrum
        (i.e. _g.p_spectrum).

    Returns
    -------
    |ndarray|
        Complex-valued array containing the FID.
    """
    p_spec = p_spec or _g.p_spectrum
    if getndim(p_spec=p_spec) != 1:
        raise ValueError("get1d_fid(): current spectrum is not 1D")
    p_fid = p_spec.parents[1] / "fid"
    fid = np.fromfile(p_fid, dtype=np.int32)
    td = fid.size
    fid = fid.reshape(int(td/2), 2)
    fid = np.transpose(fid)
    # so now fid[0] is the real part, fid[1] the imaginary
    return fid[0] + (1j * fid[1])


def _get_1d(spec_fname, bounds="", p_spec=None):
    """
    Helper-helper function which does the real work in reading the spectrum.

    The interface is the same as get1d_real, but with an additional parameter
    spec_fname. This is a string that is either "1r" (for the real spectrum) or
    "1i" (for the imaginary spectrum).
    """
    p_spec = p_spec or _g.p_spectrum
    p_specdata = p_spec / spec_fname
    spec = np.fromfile(p_specdata, dtype=np.int32)
    nc_proc = int(getpar("NC_proc", p_spec))
    spec = spec * (2 ** nc_proc)

    # Here, _g.spec_f1p and _g.spec_f2p can be ndarrays in a VERY special case.
    # The spectrum must appear to be 2D before the optimisation is started
    # (i.e. when backend.py initialises these values), and then must be changed
    # to 1D after the first acquisition. This happens, for example, in
    # dosy_opt.py. In ordinary cases, these conditionals will never occur so
    # they won't hurt.
    if isinstance(_g.spec_f1p, np.ndarray):
        _g.spec_f1p = _g.spec_f1p[1]
    if isinstance(_g.spec_f2p, np.ndarray):
        _g.spec_f2p = _g.spec_f2p[1]
    # We don't need to check if they're equal to zero. If they are, then
    # backend.py will already have set the variables in _g to be None.
    # OK now we can proceed as normal.
    if bounds == "":
        if _g.spec_f1p is None and _g.spec_f2p is None:  # DPL not used
            return spec
        else:
            # set the bounds to F1P and F2P if they are not None.
            left, right = _g.spec_f1p, _g.spec_f2p
    else:
        right, left = _parse_bounds(bounds)

    # Get default bounds
    si = int(getpar("SI", p_spec))
    left_point, right_point = 0, si - 1
    # Then replace them if necessary
    if left is not None:
        left_point = _ppm_to_point(left, p_spec=p_spec)
    if right is not None:
        right_point = _ppm_to_point(right, p_spec=p_spec)

    return spec[left_point:right_point + 1]


def get1d_real(bounds="", p_spec=None):
    """
    Return the real spectrum as a |ndarray|. This function accounts for
    TopSpin's ``NC_PROC`` variable, scaling the spectrum intensity accordingly.

    Note that this function only works for 1D spectra. It does *not* work for
    1D projections of 2D spectra. If you want to work with projections, you can
    use `get2d_rr` to get the full 2D spectrum, then manipulate it using numpy
    functions as appropriate. Examples can be found in the docs.

    The *bounds* parameter may be specified in the following formats:

     - between 5 and 8 ppm:  ``bounds="5..8"``  OR ``bounds=(5, 8)``
     - greater than 9.3 ppm: ``bounds="9.3.."`` OR ``bounds=(9.3, None)``
     - less than -2 ppm:     ``bounds="..-2"``  OR ``bounds=(None, -2)``

    Parameters
    ----------
    bounds : str or tuple, optional
        String or tuple describing the region of interest. See above for
        examples. If no bounds are provided, uses the ``F1P`` and ``F2P``
        processing parameters, which can be specified via ``dpl``. If these are
        not specified, defaults to the whole spectrum.
    p_spec : |Path|, optional
        Path to the procno folder of interest. Leave blank to use the currently
        active spectrum (i.e. _g.p_spectrum).

    Returns
    -------
    |ndarray|
        Array containing the spectrum or the desired section of it (if bounds
        were specified).
    """
    if getndim(p_spec=p_spec) != 1:
        raise ValueError("get1d_real(): current spectrum is not 1D")
    return _get_1d(spec_fname="1r", bounds=bounds, p_spec=p_spec)


def get1d_imag(bounds="", p_spec=None):
    """
    Same as `get1d_real`, except that it reads the imaginary spectrum.
    """
    if getndim(p_spec=p_spec) != 1:
        raise ValueError("get1d_imag(): current spectrum is not 1D")
    return _get_1d(spec_fname="1i", bounds=bounds, p_spec=p_spec)


def _get_2d(spec_fname, f1_bounds="", f2_bounds="", p_spec=None):
    """
    Helper-helper function which does the real work in reading the spectrum.

    The interface is the same as get2d_rr, but with an additional parameter
    spec_fname. This is a string that is either "2rr", "2ri", "2ir", or "2ii".
    """
    p_spec = p_spec or _g.p_spectrum
    p_specdata = p_spec / spec_fname

    # Check data type (TopSpin 3 int vs TopSpin 4 float)
    dtypp = getpar("dtypp", p_spec)
    if dtypp[0] == 0 and dtypp[1] == 0:  # TS3 data
        dt = "<" if np.all(getpar("bytordp", p_spec) == 0) else ">"
        dt += "i4"
    else:
        raise NotImplementedError("_get_2d: "
                                  "float data not yet accepted")
    sp = np.fromfile(p_specdata, dtype=np.dtype(dt))
    # Format according to xdim. See TopSpin "data format" manual.
    # See also http://docs.nmrfx.org/viewer/files/datasets.
    # Get si and xdim
    si = getpar("si", p_spec)
    si = (int(si[0]), int(si[1]))
    xdim = getpar("xdim", p_spec)
    xdim = (int(xdim[0]), int(xdim[1]))
    sp = sp.reshape(si)
    # Reshape.
    nrows, ncols = int(si[0]/xdim[0]), int(si[1]/xdim[1])
    submatrix_size = np.prod(xdim)
    sp = sp.reshape(si[0] * ncols, xdim[1])
    sp = np.vsplit(sp, nrows * ncols)
    sp = np.concatenate(sp, axis=1)
    sp = np.hsplit(sp, nrows)
    sp = np.concatenate(sp, axis=0)
    sp = sp.reshape(si)
    sp = sp * (2 ** getpar("nc_proc", p_spec)[1])

    # Read in DPL and overwrite bounds if the bounds were not set
    if f1_bounds == "":
        if _g.spec_f1p is not None and _g.spec_f2p is not None:  # DPL was used
            # f2p is lower than f1p.
            f1_bounds = f"{_g.spec_f2p[0]}..{_g.spec_f1p[0]}"
    if f2_bounds == "":
        if _g.spec_f1p is not None and _g.spec_f2p is not None:  # DPL was used
            f2_bounds = f"{_g.spec_f2p[1]}..{_g.spec_f1p[1]}"
    f1_lower, f1_upper = _parse_bounds(f1_bounds)
    f2_lower, f2_upper = _parse_bounds(f2_bounds)
    # Convert ppm to points
    f1_lower_point = _ppm_to_point(f1_lower, axis=0, p_spec=p_spec) \
        if f1_lower is not None else si[0] - 1
    f1_upper_point = _ppm_to_point(f1_upper, axis=0, p_spec=p_spec) \
        if f1_upper is not None else 0
    f2_lower_point = _ppm_to_point(f2_lower, axis=1, p_spec=p_spec) \
        if f2_lower is not None else si[1] - 1
    f2_upper_point = _ppm_to_point(f2_upper, axis=1, p_spec=p_spec) \
        if f2_upper is not None else 0
    return sp[f1_upper_point:f1_lower_point + 1,
              f2_upper_point:f2_lower_point + 1]


def get2d_rr(f1_bounds="", f2_bounds="", p_spec=None):
    """
    Return the real part of the 2D spectrum (the "RR" quadrant) as a 2D
    |ndarray|. This function takes into account the ``NC_PROC`` value in
    TopSpin's processing parameters.

    The *f1_bounds* and *f2_bounds* parameters may be specified in the
    following formats:

     - between 5 and 8 ppm:  ``f1_bounds="5..8"``  OR ``f1_bounds=(5, 8)``
     - greater than 9.3 ppm: ``f1_bounds="9.3.."`` OR ``f1_bounds=(9.3, None)``
     - less than -2 ppm:     ``f1_bounds="..-2"``  OR ``f1_bounds=(None, -2)``

    Parameters
    ----------
    f1_bounds : str or tuple, optional
        String or tuple describing the indirect-dimension region of interest.
        See above for examples. If no bounds are provided, uses the ``1 F1P``
        and ``1 F2P`` processing parameters, which can be specified via
        ``dpl``. If these are not specified, defaults to the whole spectrum.
    f2_bounds : str or tuple, optional
        String or tuple describing the direct-dimension region of interest. See
        above for examples. If no bounds are provided, uses the ``2 F1P`` and
        ``2 F2P`` processing parameters, which can be specified via ``dpl``. If
        these are not specified, defaults to the whole spectrum.
    p_spec : |Path|, optional
        Path to the procno folder of interest. Leave blank to use the currently
        active spectrum (i.e. _g.p_spectrum).

    Returns
    -------
    |ndarray|
        2D array containing the spectrum or the desired section of it (if
        *f1_bounds* or *f2_bounds* were specified).
    """
    if getndim(p_spec=p_spec) != 2:
        raise ValueError("get2d_rr(): current spectrum is not 2D")
    return _get_2d(spec_fname="2rr",
                   f1_bounds=f1_bounds,
                   f2_bounds=f2_bounds,
                   p_spec=p_spec)


def get2d_ri(f1_bounds="", f2_bounds="", p_spec=None):
    """
    Same as `get2d_rr`, except that it reads the '2ri' file.
    """
    if getndim(p_spec=p_spec) != 2:
        raise ValueError("get2d_ri(): current spectrum is not 2D")
    return _get_2d(spec_fname="2ri",
                   f1_bounds=f1_bounds,
                   f2_bounds=f2_bounds,
                   p_spec=p_spec)


def get2d_ir(f1_bounds="", f2_bounds="", p_spec=None):
    """
    Same as `get2d_rr`, except that it reads the '2ir' file.
    """
    if getndim(p_spec=p_spec) != 2:
        raise ValueError("get2d_ir(): current spectrum is not 2D")
    return _get_2d(spec_fname="2ir",
                   f1_bounds=f1_bounds,
                   f2_bounds=f2_bounds,
                   p_spec=p_spec)


def get2d_ii(f1_bounds="", f2_bounds="", p_spec=None):
    """
    Same as `get2d_rr`, except that it reads the '2ii' file.
    """
    if getndim(p_spec=p_spec) != 2:
        raise ValueError("get2d_ii(): current spectrum is not 2D")
    return _get_2d(spec_fname="2ii",
                   f1_bounds=f1_bounds,
                   f2_bounds=f2_bounds,
                   p_spec=p_spec)


def _get_acqu_par(par, p_acqus):
    """
    Obtains the value of an acquisition parameter.

    Note that pulse powers in dB (PLdB / SPdB) cannot be obtained using this
    function, as they are not stored in the acqus file.

    Parameters
    ----------
    par : str
        Name of the acquisition parameter.
    p_acqus : |Path|
        Path to the status acquisition file (this is 'acqus' for 1D spectra and
        direct dimension of 2D spectra, or 'acqu2s' for indirect dimension of
        2D spectra).

    Returns
    -------
    float or None
        Value of the acquisition parameter. None if the value is not a number,
        or if the parameter doesn't exist.
    """
    # Capitalise and remove any spaces from par
    par = par.upper()
    if len(par.split()) > 1:
        par = "".join(par.split())

    # Split par into number-less bit and number bit
    parl = par.rstrip("1234567890")
    parr = par[len(parl):]
    list_params = ("AMP AMPCOIL BWFAC CAGPARS CNST CPDPRG D FCUCHAN"
                   " FN_INDIRECT FS GPNAM GPX GPY GPZ HGAIN HPMOD IN INF INP"
                   " INTEGFAC L MULEXPNO P PACOIL PCPD PEXSEL PHCOR PL PLW"
                   " PLWMAX PRECHAN PROBINPUTS RECCHAN RECPRE RECPRFX RECSEL"
                   " RSEL S SELREC SP SPNAM SPOAL SPOFFS SPPEX SPW SUBNAM"
                   " SWIBOX TD_INDIRECT TE_STAB TL TOTROT XGAIN").split()
    # Get the parameter
    if (parr != "") and (parl in list_params):  # e.g. cnst2
        with open(p_acqus, "r") as file:
            # Read up to the line declaring the parameters
            for line in file:
                if line.upper().startswith(f"##${parl}="):
                    break
            else:   # triggers if didn't break -- i.e. parameter was not found
                return None
            # Grab the values and put them in a list
            s = ""
            # Read until next parameter
            line = file.readline()
            while line != "" and not line.startswith("##"):
                s = s + line + " "
                line = file.readline()
            # Pick out the desired value and return it
            value = s.split()[int(parr)]
            try:
                return float(value)
            except ValueError:  # not a float
                return None
    else:                                             # e.g. sfo1 or rg
        with open(p_acqus, "r") as file:
            for line in file:
                if line.upper().startswith(f"##${par}="):
                    value = line.split(maxsplit=1)[-1].strip()
                    # strip away surrounding angle brackets
                    if value[0] == '<' and value[-1] == '>':
                        value = value[1:-1]
                    try:
                        return float(value)
                    except ValueError:  # not a float
                        return None
    # If it hasn't been found
    return None


def _get_proc_par(par, p_procs):
    """
    Obtains the value of a processing parameter.

    Parameters
    ----------
    par : str
        Name of the processing parameter.
    p_procs : |Path|
        Path to the status processing file (this is 'procs' for 1D spectra and
        direct dimension of 2D spectra, or 'proc2s' for indirect dimension of
        2D spectra).

    Returns
    -------
    float or None
        Value of the processing parameter. None if the value is not a number,
        or if the parameter doesn't exist.
    """
    # Capitalise and remove any spaces from par
    par = par.upper()
    if len(par.split()) > 1:
        par = "".join(par.split())
    # Get the value (for processing parameters there aren't any lists like
    # CNST/D/P)
    with open(p_procs, "r") as file:
        for line in file:
            if line.upper().startswith(f"##${par}="):
                value = line.split(maxsplit=1)[-1].strip()
                # strip away surrounding angle brackets
                if value[0] == '<' and value[-1] == '>':
                    value = value[1:-1]
                try:
                    return float(value)
                except ValueError:
                    return None


def getpar(par, p_spec=None):
    """
    Obtains the value of a numeric (acquisition or processing) parameter.
    Non-numeric parameters (i.e. strings) are not currently accessible! Works
    for both 1D and 2D spectra (see return type below), but nothing higher.

    Parameters
    ----------
    par : str
        Name of the parameter.
    p_spec : |Path|, optional
        Path to the procno folder of interest. Leave blank to use the currently
        active spectrum (i.e. _g.p_spectrum).

    Returns
    -------
    float or |ndarray|
        Value(s) of the requested parameter. None if the given parameter was
        not found.

        For parameters that exist for both dimensions of 2D spectra, getpar()
        returns an ndarray consisting of (f1_value, f2_value).  Otherwise (for
        1D spectra, or for 2D parameters which only apply to the direct
        dimension), getpar() returns a float.
    """
    p_spec = p_spec or _g.p_spectrum

    # Figure out the number of dimensions.
    ndim = getndim(p_spec=p_spec)
    if ndim == 1:
        acq = _get_acqu_par(par, p_spec.parents[1] / "acqus")
        if acq is not None:
            return acq
        else:
            return _get_proc_par(par, p_spec / "procs")
    elif ndim == 2:
        # Try to get acquisition parameters first.
        acq_f1 = _get_acqu_par(par, p_spec.parents[1] / "acqu2s")
        acq_f2 = _get_acqu_par(par, p_spec.parents[1] / "acqus")
        # If both were found, return them as an nparray
        if acq_f1 is not None and acq_f2 is not None:
            return np.array([acq_f1, acq_f2])
        # If only f2 value was found, return it
        elif acq_f2 is not None:
            return acq_f2
        # If reached here, then means that acquisition parameter was not found.
        # Try to get processing parameters.
        proc_f1 = _get_proc_par(par, p_spec / "proc2s")
        proc_f2 = _get_proc_par(par, p_spec / "procs")
        # If both were found, return them as an nparray
        if proc_f1 is not None and proc_f2 is not None:
            return np.array([proc_f1, proc_f2])
        # If only f2 value was found, return it. In fact, based on the test
        # data, this will never be the case: every parameter inside procs is
        # also found inside proc2s. So this code is technically unreachable.
        elif proc_f2 is not None:
            return proc_f2
        # If reached here, neither were found, so return None.
        return None
    else:
        raise ValueError("getpar() only works for 1D and 2D spectra.")


def getndim(p_spec=None):
    """
    Obtains the dimensionality of the spectrum, i.e. the status value of
    PARMODE. Note that Bruker uses PARMODE=n for (n+1)D spectra, whereas this
    function simply returns (n+1) (as an int).

    Note that we call _get_acqu_par() instead of getpar() to avoid an infinite
    loop.

    Parameters
    ----------
    p_spec : |Path|, optional
        Path to the procno folder of interest. Leave blank to use the currently
        active spectrum (i.e. _g.p_spectrum).

    Returns
    -------
    int
        Dimensionality of the spectrum.
    """
    p_spec = p_spec or _g.p_spectrum
    p_acqus = p_spec.parents[1] / "acqus"
    # for (n+1)D, bruker_ndim = n
    bruker_ndim = _get_acqu_par("PARMODE", p_acqus)
    return int(bruker_ndim) + 1


def getnfev():
    """
    Returns the number of NMR spectra evaluated so far. This will be equal to 1
    (not 0) the first time the cost function is called, since the cost function
    is only called after the NMR spectrum has been acquired.
    """
    return _g.nfev


def log(s):
    """
    Prints a string to the poise.log file. If this is called from inside a cost
    function, the text is printed *before* the cost function is evaluated, so
    will appear above the corresponding function evaluation.

    Parameters
    ----------
    s : object
        The object to be printed. Typically a string, but since this is just
        passed directly to print(), anything with a __str__() method can be
        used.

    Returns
    -------
    None
    """
    with open(_g.p_optlog, 'a') as fp:
        print(s, file=fp)
